import json
import os

import docopt
import pkg_resources

import dcoscli
from dcos import cmds, emitting, http, options, package, subcommand, util
from dcos.errors import DCOSException
from dcos.package import get_package_manager
from dcoscli import tables
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import confirm, confirm_text, decorate_docopt_usage

logger = util.get_logger(__name__)

emitter = emitting.FlatEmitter()


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("package"),
        argv=argv,
        version='dcos-package version {}'.format(dcoscli.version))
    http.silence_requests_warnings()

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: All of the supported commands
    :rtype: dcos.cmds.Command
    """

    return [
        cmds.Command(
            hierarchy=['package', 'repo', 'list'],
            arg_keys=['--json'],
            function=_list_repos),

        cmds.Command(
            hierarchy=['package', 'repo', 'add'],
            arg_keys=['<repo-name>', '<repo-url>', '--index'],
            function=_add_repo),

        cmds.Command(
            hierarchy=['package', 'repo', 'import'],
            arg_keys=['<repos-file>'],
            function=_import_repos),

        cmds.Command(
            hierarchy=['package', 'repo', 'remove'],
            arg_keys=['<repo-names>'],
            function=_remove_repo),

        cmds.Command(
            hierarchy=['package', 'describe'],
            arg_keys=['<package-name>', '--app', '--cli', '--options',
                      '--render', '--package-versions', '--package-version',
                      '--config'],
            function=_describe),

        cmds.Command(
            hierarchy=['package', 'install'],
            arg_keys=['<package-name>', '--package-version', '--options',
                      '--app-id', '--cli', '--global', '--app', '--yes'],
            function=_install),

        cmds.Command(
            hierarchy=['package', 'list'],
            arg_keys=['--json', '--app-id', '--cli', '<package-name>'],
            function=_list),

        cmds.Command(
            hierarchy=['package', 'search'],
            arg_keys=['--json', '<query>'],
            function=_search),

        cmds.Command(
            hierarchy=['package', 'uninstall'],
            arg_keys=['<package-name>', '--all', '--app-id', '--cli', '--app',
                      '--yes'],
            function=_uninstall),

        cmds.Command(
            hierarchy=['package'],
            arg_keys=['--config-schema', '--info'],
            function=_package),
    ]


def _package(config_schema, info):
    """
    :param config_schema: Whether to output the config schema
    :type config_schema: boolean
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: Process status
    :rtype: int
    """
    if config_schema:
        schema = json.loads(
            pkg_resources.resource_string(
                'dcos',
                'data/config-schema/package.json').decode('utf-8'))
        emitter.publish(schema)
    elif info:
        emitter.publish(default_command_info("package"))
    else:
        doc = default_doc("package")
        emitter.publish(options.make_generic_usage_message(doc))
        return 1

    return 0


def _list_repos(is_json):
    """List configured package repositories.

    :param is_json: output json if True
    :type is_json: bool
    :returns: Process status
    :rtype: int
    """

    package_manager = get_package_manager()
    repos = package_manager.get_repos()

    if is_json:
        return emitter.publish(repos)
    elif repos.get("repositories"):
        repos = ["{}: {}".format(repo.get("name"), repo.get("uri"))
                 for repo in repos.get("repositories")]
        emitter.publish("\n".join(repos))
    else:
        msg = ("There are currently no repos configured. "
               "Please use `dcos package repo add` to add a repo")
        raise DCOSException(msg)

    return 0


def _add_repo(repo_name, repo_url, index):
    """Add package repo and update repo with new repo

    :param repo_name: name to call repo
    :type repo_name: str
    :param repo_url: location of repo to add
    :type repo_url: str
    :param index: index to add this repo
    :type index: int
    :rtype: None
    """

    package_manager = get_package_manager()
    package_manager.add_repo(repo_name, repo_url, index)

    return 0


def _import_repos(repos_file):
    """Add package repos from a JSON file

    :param repos_file: path to JSON file containing repos.
    :type repos_file: str
    :rtype: int
    """

    package_manager = get_package_manager()

    with open(repos_file) as f:
        try:
            data = json.load(f)
        except json.JSONDecodeError:
            _raise_invalid_repos_file()
        repositories = data.get('repositories')
        if not repositories:
            _raise_invalid_repos_file()

        for index, repo in enumerate(data['repositories']):
            repo_name = repo.get('name')
            repo_uri = repo.get('uri')
            if not repo_name or not repo_uri:
                emitter.publish('Repo missing name or uri. Skipping.')
                continue
            try:
                package_manager.add_repo(repo['name'], repo['uri'], index)
                emitter.publish('Added repo "%s" (%s) at index %d'
                                % (repo_name, repo_uri, index))
            except DCOSException as e:
                emitter.publish('Error (%s) while adding repo "%s" (%s). '
                                'Skipping.'
                                % (e, repo_name, repo_uri))
                continue

    return 0


def _raise_invalid_repos_file():
    raise DCOSException(
        'No repositories found to import. '
        'You should provide a file containing package '
        'repositories, listed in the format of `dcos package list --json`.'
        '\nExample: '
        '{"repositories": [{"name": "Universe", "uri": "uri-here"}]}')


def _remove_repo(repo_names):
    """Remove package repo and update repo with new repo

    :param repo_names: names of repos
    :type repo_name: [str]
    :returns: Process status
    :rtype: int
    """

    package_manager = get_package_manager()
    for repo_name in repo_names:
        package_manager.remove_repo(repo_name)

    return 0


def _describe(package_name,
              app,
              cli,
              options_path,
              render,
              package_versions,
              package_version,
              config):
    """Describe the specified package.

    :param package_name: The package to describe
    :type package_name: str
    :param app: If True, marathon.json will be printed
    :type app: boolean
    :param cli: If True, command.json | resource.json's cli property should
                be printed
    :type cli: boolean
    :param options_path: Path to json file with options to override
                         config.json defaults.
    :type options_path: str
    :param render: If True, marathon.json will be rendered
    :type render: boolean
    :param package_versions: If True, a list of all package versions will
                             be printed
    :type package_versions: boolean
    :param package_version: package version
    :type package_version: str | None
    :param config: If True, config.json will be printed
    :type config: boolean
    :returns: Process status
    :rtype: int
    """

    # If the user supplied template options, they definitely want to
    # render the template
    if options_path:
        render = True

    # Fail early if options file isn't valid
    user_options = util.read_file_json(options_path)

    package_manager = get_package_manager()
    pkg = package_manager.get_package_version(package_name, package_version)

    pkg_json = pkg.package_json()

    if package_versions:
        emitter.publish(pkg.package_versions())
    elif cli or app or config:
        if cli:
            emitter.publish(pkg.cli_definition())
        if app:
            if render:
                app_output = pkg.marathon_json(user_options)
            else:
                app_output = pkg.marathon_template().decode("UTF-8")
                if app_output and app_output[-1] == '\n':
                    app_output = app_output[:-1]
            emitter.publish(app_output)
        if config:
            config_output = pkg.config_json()
            emitter.publish(config_output)
    else:
        emitter.publish(pkg_json)

    return 0


def _install(package_name, package_version, options_path, app_id, cli,
             global_, app, yes):
    """Install the specified package.

    :param package_name: the package to install
    :type package_name: str
    :param package_version: package version to install
    :type package_version: str
    :param options_path: path to file containing option values
    :type options_path: str
    :param app_id: app ID for installation of this package
    :type app_id: str
    :param cli: indicates if the cli should be installed
    :type cli: bool
    :param global_: indicates that the cli should be installed globally
    :type global_: bool
    :param app: indicate if the application should be installed
    :type app: bool
    :param yes: automatically assume yes to all prompts
    :type yes: bool
    :returns: process status
    :rtype: int
    """

    if cli is False and app is False:
        # Install both if neither flag is specified
        cli = app = True

    # Fail early if options file isn't valid
    user_options = util.read_file_json(options_path)

    package_manager = get_package_manager()
    pkg = package_manager.get_package_version(package_name, package_version)

    pkg_json = pkg.package_json()

    selected = pkg_json.get('selected')
    if selected:
        link = ('https://mesosphere.com/'
                'catalog-terms-conditions/#certified-services')
    else:
        link = ('https://mesosphere.com/'
                'catalog-terms-conditions/#community-services')
    emitter.publish(
        ('By Deploying, you agree to '
         'the Terms and Conditions ' + link)
    )

    pre_install_notes = pkg_json.get('preInstallNotes')
    if app and pre_install_notes:
        emitter.publish(pre_install_notes)

    if not confirm('Continue installing?', yes):
        emitter.publish('Exiting installation.')
        return 0

    if app and pkg.marathon_template():

        # Even though package installation will check for template rendering
        # errors, we want to fail early, before trying to install.
        pkg.options(user_options)

        # Install in Marathon
        msg = 'Installing Marathon app for package [{}] version [{}]'.format(
            pkg.name(), pkg.version())
        if app_id is not None:
            msg += ' with app id [{}]'.format(app_id)

        emitter.publish(msg)

        package_manager.install_app(pkg, user_options, app_id)

    if cli and pkg.cli_definition():
        # Install subcommand
        msg = 'Installing CLI subcommand for package [{}] version [{}]'.format(
            pkg.name(), pkg.version())
        emitter.publish(msg)

        subcommand.install(pkg, global_)

        subcommand_paths = subcommand.get_package_commands(package_name)
        new_commands = [os.path.basename(p).replace('-', ' ', 1)
                        for p in subcommand_paths]

        if new_commands:
            commands = ', '.join(new_commands)
            plural = "s" if len(new_commands) > 1 else ""
            emitter.publish("New command{} available: {}".format(plural,
                                                                 commands))

    post_install_notes = pkg_json.get('postInstallNotes')
    if app and post_install_notes:
        emitter.publish(post_install_notes)

    return 0


def _list(json_, app_id, cli_only, package_name):
    """List installed apps

    :param json_: output json if True
    :type json_: bool
    :param app_id: App ID of app to show
    :type app_id: str
    :param cli_only: if True, only show packages with installed subcommands
    :type cli_only: bool
    :param package_name: The package to show
    :type package_name: str
    :returns: process return code
    :rtype: int
    """

    package_manager = get_package_manager()
    if app_id is not None:
        app_id = util.normalize_marathon_id_path(app_id)
    results = package.installed_packages(
        package_manager, app_id, package_name, cli_only)

    # only emit those packages that match the provided package_name and app_id
    if results or json_:
        emitting.publish_table(emitter, results, tables.package_table, json_)
    else:
        msg = ("There are currently no installed packages. "
               "Please use `dcos package install` to install a package.")
        raise DCOSException(msg)
    return 0


def _matches_package_name(name, pkg_info):
    """
    :param name: the name of the package
    :type name: str
    :param pkg_info: the package description
    :type pkg_info: dict
    :returns: True if the name is not defined or the package matches that name;
              False otherwise
    :rtype: bool
    """

    return name is None or pkg_info['name'] == name


def _matches_app_id(app_id, pkg_info):
    """
    :param app_id: the application id
    :type app_id: str
    :param pkg_info: the package description
    :type pkg_info: dict
    :returns: True if the app id is not defined or the package matches that app
              id; False otherwize
    :rtype: bool
    """

    return app_id is None or app_id in pkg_info.get('apps')


def _search(json_, query):
    """Search for matching packages.

    :param json_: output json if True
    :type json_: bool
    :param query: The search term
    :type query: str
    :returns: Process status
    :rtype: int
    """

    if not query:
        query = ''

    package_manager = get_package_manager()
    results = package_manager.search_sources(query)

    if json_ or results['packages']:
        emitting.publish_table(emitter,
                               results,
                               tables.package_search_table,
                               json_)
    else:
        raise DCOSException('No packages found.')
    return 0


def _uninstall(package_name, remove_all, app_id, cli, app, skip_confirmation):
    """Uninstall the specified package.

    :param package_name: The package to uninstall
    :type package_name: str
    :param remove_all: Whether to remove all instances of the named package
    :type remove_all: boolean
    :param app_id: App ID of the package instance to uninstall
    :type app_id: str
    :param skip_confirmation: Skip confirmation for uninstall
    :type skip_confirmation: boolean
    :returns: Process status
    :rtype: int
    """

    # Don't gate CLI uninstalls.
    # Don't gate uninstalls if the user wants to skip confirmation.
    if not cli and not skip_confirmation and not _confirm_uninstall(
        package_name,
        remove_all,
        app_id
    ):
        return 0

    package_manager = get_package_manager()
    err = package.uninstall(
        package_manager, package_name, remove_all, app_id, cli, app)
    if err is not None:
        emitter.publish(err)
        return 1

    return 0


def _confirm_uninstall(package_name, remove_all, app_id):
    """Interactively confirm the uninstall with the user.

    :param package_name: The package to uninstall
    :type package_name: str
    :param remove_all: Whether to remove all instances of the named package
    :type remove_all: boolean
    :param app_id: App ID of the package instance to uninstall
    :type app_id: str
    :returns: Whether the uninstall should proceeed or not.
    :rtype: bool
    """

    if remove_all:
        confirm_cta = ("WARNING: This action cannot be undone. This will "
                       "uninstall all instances of the [{package_name}] "
                       "package, and delete all data for all {package_name} "
                       "instances.").format(package_name=package_name)
        confirm_prompt = "Please type 'uninstall all {}'".format(package_name)
        expected_text = "uninstall all {}".format(package_name)
    elif app_id:
        confirm_cta = ("WARNING: This action cannot be undone. This will "
                       "uninstall [{}] and delete all of its persistent "
                       "(logs, configurations, database artifacts, "
                       "everything).").format(app_id)
        confirm_prompt = "Please type the name of the service to confirm"
        expected_text = app_id
    else:
        confirm_cta = ("WARNING: This action cannot be undone. This will "
                       "uninstall [{}] and delete all of its persistent "
                       "data (logs, configurations, database artifacts, "
                       "everything).").format(package_name)
        confirm_prompt = "Please type the name of the service to confirm"
        expected_text = package_name

    emitter.publish(confirm_cta)
    if not confirm_text(confirm_prompt, expected_text):
        emitter.publish("Cancelling uninstall.")
        return False
    else:
        return True



import re

import jsonschema
import mock
import pytest
import requests
from requests.structures import CaseInsensitiveDict

from dcos import http, marathon, rpcclient
from dcos.errors import DCOSException, DCOSHTTPException


def test_add_pod_puts_json_in_request_body():
    _assert_add_pod_puts_json_in_request_body(pod_json={"some": "json"})
    _assert_add_pod_puts_json_in_request_body(
        pod_json=["another", {"json": "value"}])


def test_add_pod_returns_parsed_response_body():
    _assert_add_pod_returns_parsed_response_body({"id": "i-am-a-pod"})
    _assert_add_pod_returns_parsed_response_body(["another", "pod", "json"])


def test_remove_pod_has_default_force_value():
    marathon_client, rpc_client = _create_fixtures()
    marathon_client.remove_pod('foo')
    rpc_client.http_req.assert_called_with(
        http.delete, 'v2/pods/foo', params=None)


def test_remove_pod_builds_rpc_correctly():
    _assert_remove_pod_builds_rpc_correctly(
        pod_id='foo', force=False, path='v2/pods/foo', params=None)
    _assert_remove_pod_builds_rpc_correctly(
        pod_id='foo', force=True, path='v2/pods/foo', params={'force': 'true'})
    _assert_remove_pod_builds_rpc_correctly(
        pod_id='/foo bar/', force=False, path='v2/pods/foo%20bar', params=None)


def test_remove_pod_propagates_dcos_exception():
    _assert_method_propagates_rpc_dcos_exception(
        lambda marathon_client: marathon_client.remove_pod('bad'))


def test_show_pod_builds_rpc_correctly():
    _assert_show_pod_builds_rpc_correctly(
        pod_id='foo', path='v2/pods/foo::status')
    _assert_show_pod_builds_rpc_correctly(
        pod_id='/bar', path='v2/pods/bar::status')
    _assert_show_pod_builds_rpc_correctly(
        pod_id='baz/', path='v2/pods/baz::status')
    _assert_show_pod_builds_rpc_correctly(
        pod_id='foo bar', path='v2/pods/foo%20bar::status')


def test_show_pod_returns_response_json():
    _assert_show_pod_returns_response_json({'some': 'json'})
    _assert_show_pod_returns_response_json(['another', 'json', 'value'])


def test_show_pod_propagates_dcos_exception():
    _assert_method_propagates_rpc_dcos_exception(
        lambda marathon_client: marathon_client.show_pod('bad-req'))


def test_show_pod_raises_dcos_exception_for_json_parse_errors():
    _assert_method_raises_dcos_exception_for_json_parse_errors(
        lambda marathon_client: marathon_client.show_pod('bad-json'))


def test_list_pod_builds_rpc_correctly():
    marathon_client, rpc_client = _create_fixtures()
    marathon_client.list_pod()
    rpc_client.http_req.assert_called_with(http.get, 'v2/pods/::status')


def test_list_pod_returns_success_response_json():
    _assert_list_pod_returns_success_response_json(body_json={'some': 'json'})
    _assert_list_pod_returns_success_response_json(body_json=['a', 'b', 'c'])


def test_list_pod_propagates_rpc_dcos_exception():
    _assert_method_propagates_rpc_dcos_exception(
        lambda marathon_client: marathon_client.list_pod())


def test_list_pod_raises_dcos_exception_for_json_parse_errors():
    _assert_method_raises_dcos_exception_for_json_parse_errors(
        lambda marathon_client: marathon_client.list_pod())


def test_update_pod_executes_successfully():
    _assert_update_pod_executes_successfully(
        pod_id='foo',
        pod_json={'some', 'json'},
        force=False,
        path='v2/pods/foo',
        params=None,
        deployment_id='pod-deployment-id')
    _assert_update_pod_executes_successfully(
        pod_id='/foo bar/',
        pod_json={'some', 'json'},
        force=False,
        path='v2/pods/foo%20bar',
        params=None,
        deployment_id='pod-deployment-id')
    _assert_update_pod_executes_successfully(
        pod_id='foo',
        pod_json={'some', 'json'},
        force=True,
        path='v2/pods/foo',
        params={'force': 'true'},
        deployment_id='pod-deployment-id')
    _assert_update_pod_executes_successfully(
        pod_id='foo',
        pod_json={'something', 'different'},
        force=False,
        path='v2/pods/foo',
        params=None,
        deployment_id='pod-deployment-id')
    _assert_update_pod_executes_successfully(
        pod_id='foo',
        pod_json={'some', 'json'},
        force=False,
        path='v2/pods/foo',
        params=None,
        deployment_id='an-arbitrary-value')


def test_update_pod_has_default_force_value():
    marathon_client, rpc_client = _create_fixtures()
    marathon_client.update_pod('foo', {'some': 'json'})
    rpc_client.http_req.assert_called_with(
        http.put, 'v2/pods/foo', params=None, json={'some': 'json'})


def test_update_pod_propagates_rpc_dcos_exception():
    _assert_method_propagates_rpc_dcos_exception(
        lambda marathon_client:
            marathon_client.update_pod('foo', {'some': 'json'}))


def test_add_pod_raises_dcos_exception_if_deployment_id_missing():
    _assert_add_pod_raises_dcos_exception_if_deployment_id_missing(
        headers={'foo': 'bar'})
    _assert_add_pod_raises_dcos_exception_if_deployment_id_missing(
        headers={'marathon-deployment_ID': 'misspelled-field', 'zzz': 'aaa'})


def test_update_pod_raises_dcos_exception_if_deployment_id_missing():
    _assert_update_pod_raises_dcos_exception_if_deployment_id_missing(
        headers={'foo': 'bar'})
    _assert_update_pod_raises_dcos_exception_if_deployment_id_missing(
        headers={'marathon-deployment_ID': 'misspelled-field', 'zzz': 'aaa'})


def test_kill_pod_instances_executes_successfully():
    pod_id = 'foo'
    instance_ids = ['instance1', 'instance2']
    path = 'v2/pods/foo::instances'
    response_json = {'some': ['instance', 'status']}

    mock_response = mock.create_autospec(requests.Response)
    mock_response.json.return_value = response_json

    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.return_value = mock_response

    path_format_method_name = 'dcos.marathon.Client._marathon_id_path_format'
    path_format = mock.MagicMock()
    path_format.return_value = path

    with mock.patch(path_format_method_name, new=path_format):
        actual_json = marathon_client.kill_pod_instances(pod_id, instance_ids)

    path_format.assert_called_with('v2/pods/{}::instances', pod_id)
    rpc_client.http_req.assert_called_with(
        http.delete, path, json=instance_ids)
    assert actual_json == response_json


def test_kill_pod_instances_propagates_rpc_dcos_exception():
    _assert_method_propagates_rpc_dcos_exception(
        lambda marathon_client:
            marathon_client.kill_pod_instances('foo', ['inst1', 'inst2']))


def test_kill_pod_instances_raises_dcos_exception_for_json_parse_errors():
    _assert_method_raises_dcos_exception_for_json_parse_errors(
        lambda marathon_client: marathon_client.kill_pod_instances(
            'foo', ['bar', 'baz']))


@mock.patch('dcos.http.head')
def test_pod_feature_supported_gets_success_response(head_fn):
    def invoke_test_case(status_code):
        mock_response = mock.create_autospec(requests.Response)
        mock_response.status_code = status_code
        head_fn.return_value = mock_response

        rpc_client = rpcclient.RpcClient('http://base/url', timeout=42)
        marathon_client = marathon.Client(rpc_client)
        is_supported = marathon_client.pod_feature_supported()

        head_fn.assert_called_with('http://base/url/v2/pods', timeout=42)

        return is_supported

    assert invoke_test_case(status_code=200)
    assert invoke_test_case(status_code=204)

    assert not invoke_test_case(status_code=100)
    assert not invoke_test_case(status_code=302)


@mock.patch('dcos.http.head')
def test_pod_feature_supported_gets_404_response(head_fn):
    mock_response = mock.create_autospec(requests.Response)
    mock_response.status_code = 404
    head_fn.side_effect = DCOSHTTPException(mock_response)

    rpc_client = rpcclient.RpcClient('http://base/url', timeout=24)
    marathon_client = marathon.Client(rpc_client)

    assert not marathon_client.pod_feature_supported()
    head_fn.assert_called_with('http://base/url/v2/pods', timeout=24)


def test_pod_feature_supported_converts_http_exceptions_to_dcos_exceptions():
    @mock.patch('dcos.http.head')
    def test_case(head_fn, status_code):
        request = requests.Request(method='ANY', url='http://arbitrary/url')
        mock_response = mock.create_autospec(requests.Response)
        mock_response.status_code = status_code
        mock_response.reason = 'Arbitrary Reason'
        mock_response.request = request
        mock_response.json.side_effect = ValueError('empty body')
        head_fn.side_effect = DCOSHTTPException(mock_response)

        rpc_client = rpcclient.RpcClient('http://does/not/matter')
        marathon_client = marathon.Client(rpc_client)

        with pytest.raises(DCOSException) as exception_info:
            marathon_client.pod_feature_supported()

        message = rpcclient.RpcClient.response_error_message(
            status_code,
            reason='Arbitrary Reason',
            request_method='ANY',
            request_url='http://arbitrary/url',
            json_body=None)
        assert str(exception_info.value).endswith(message)

    test_case(status_code=400)
    test_case(status_code=401)
    test_case(status_code=403)
    test_case(status_code=409)
    test_case(status_code=422)
    test_case(status_code=500)


def test_pod_feature_supported_propagates_other_exceptions():
    _assert_pod_feature_supported_raises_exception(
        exception=DCOSException("BOOM!"))
    _assert_pod_feature_supported_raises_exception(
        exception=Exception("Uh oh"))


def test_rpc_client_http_req_calls_method_fn():
    def test_case(base_url, path, full_url):
        method_fn = mock.Mock()

        rpc_client = rpcclient.RpcClient(base_url)
        rpc_client.http_req(method_fn, path)

        method_fn.assert_called_with(full_url,
                                     timeout=http.DEFAULT_TIMEOUT)

    test_case(
        base_url='http://base/url',
        path='some/path',
        full_url='http://base/url/some/path')
    test_case(
        base_url='http://base/url',
        path='different/thing',
        full_url='http://base/url/different/thing')
    test_case(
        base_url='gopher://different/thing',
        path='some/path',
        full_url='gopher://different/thing/some/path')
    test_case(
        base_url='http://base/without/slash',
        path='/path/with/slash',
        full_url='http://base/without/slash/path/with/slash')
    test_case(
        base_url='http://base/with/slash/',
        path='path/without/slash',
        full_url='http://base/with/slash/path/without/slash')
    test_case(
        base_url='http://base/with/slash/',
        path='/path/with/slash',
        full_url='http://base/with/slash/path/with/slash')


def test_rpc_client_http_req_passes_args_to_method_fn():
    method_fn = mock.Mock()

    rpc_client = rpcclient.RpcClient('http://base/url')
    rpc_client.http_req(method_fn, 'some/path', 'foo', 42)

    method_fn.assert_called_with('http://base/url/some/path',
                                 'foo',
                                 42,
                                 timeout=http.DEFAULT_TIMEOUT)


def test_rpc_client_http_req_passes_kwargs_to_method_fn():
    method_fn = mock.Mock()

    rpc_client = rpcclient.RpcClient('http://base/url')
    rpc_client.http_req(method_fn, 'some/path', foo='bar', baz=42)

    method_fn.assert_called_with('http://base/url/some/path',
                                 foo='bar',
                                 baz=42,
                                 timeout=http.DEFAULT_TIMEOUT)


def test_rpc_client_http_req_kwarg_timeout_overrides_default():
    method_fn = mock.Mock()

    rpc_client = rpcclient.RpcClient('http://base/url')
    rpc_client.http_req(method_fn, 'some/path', timeout=42)

    method_fn.assert_called_with('http://base/url/some/path', timeout=42)


def test_rpc_client_http_req_set_timeout_in_constructor():
    method_fn = mock.Mock()

    rpc_client = rpcclient.RpcClient('http://base/url', 24)
    rpc_client.http_req(method_fn, 'some/path')

    method_fn.assert_called_with('http://base/url/some/path', timeout=24)


def test_rpc_client_http_req_returns_method_fn_result():
    def test_case(expected):

        def method_fn(*args, **kwargs):
            return expected

        rpc_client = rpcclient.RpcClient('http://base/url')
        actual = rpc_client.http_req(method_fn, 'some/path')

        assert actual == expected

    test_case(['the', 'result'])
    test_case({'another': 'result'})


def test_rpc_client_http_req_propagates_method_fn_exception_1():
    request = requests.Request(method='ANY', url='http://arbitrary/url')
    # Need the mock so that the `.json()` method can be overridden
    response = mock.create_autospec(requests.Response)
    response.status_code = 403
    response.reason = 'Forbidden'
    response.request = request
    response.json.side_effect = Exception('not JSON')

    def method_fn(*args, **kwargs):
        raise DCOSHTTPException(response)

    rpc_client = rpcclient.RpcClient('http://base/url')
    with pytest.raises(DCOSException) as e:
        rpc_client.http_req(method_fn, 'some/path')

    expected_message = rpcclient.RpcClient.response_error_message(
        status_code=403,
        reason='Forbidden',
        request_method='ANY',
        request_url='http://arbitrary/url',
        json_body=None)
    assert str(e).endswith(expected_message)


def test_rpc_client_http_req_propagates_method_fn_exception_2():
    request = requests.Request(method='NONE', url='http://host/path')
    # Need the mock so that the `.json()` method can be overridden
    response = mock.create_autospec(requests.Response)
    response.status_code = 422
    response.reason = 'Something Bad'
    response.request = request
    response.json.return_value = {'message': 'BOOM!'}

    def method_fn(*args, **kwargs):
        raise DCOSHTTPException(response)

    rpc_client = rpcclient.RpcClient('http://base/url')
    with pytest.raises(DCOSException) as e:
        rpc_client.http_req(method_fn, 'some/path')

    expected_message = rpcclient.RpcClient.response_error_message(
        status_code=422,
        reason='Something Bad',
        request_method='None',
        request_url='http://host/path',
        json_body={'message': 'BOOM!'})
    assert str(e).endswith(expected_message)


def test_error_json_schema_is_valid():
    error_json_schema = rpcclient.load_error_json_schema()
    jsonschema.Draft4Validator.check_schema(error_json_schema)


def test_response_error_message_with_400_status_no_json():
    _assert_response_error_message_with_400_status_no_json(
        reason=_REASON_1,
        request_method=_METHOD_1,
        request_url=_URL_1)

    _assert_response_error_message_with_400_status_no_json(
        reason=_REASON_2,
        request_method=_METHOD_2,
        request_url=_URL_2)


def test_response_error_message_with_400_status_json():
    _assert_response_error_message_with_400_status_json(
        _RESPONSE_JSON_1, _PRINTED_JSON_1)

    _assert_response_error_message_with_400_status_json(
        _RESPONSE_JSON_2, _PRINTED_JSON_2)


def test_res_err_msg_with_409_status():
    def test_case(request_url, expected_resource):
        actual = rpcclient.RpcClient.response_error_message(
            status_code=409,
            reason=_REASON_X,
            request_method=_METHOD_X,
            request_url=request_url,
            json_body=None)

        pattern = r'Changes blocked: deployment already in progress for (.*)\.'
        _assert_matches_with_groups(pattern, actual, (expected_resource,))

    test_case('http://marathon/v2/apps', 'app')
    test_case('http://marathon/v2/groups/group-id', 'group')
    test_case('http://marathon/v2/pods/', 'pod')
    test_case('http://marathon/v2/thingies/foo', 'resource')
    test_case('http://dcos/service/marathon/v2/apps/bar', 'app')
    test_case('http://pods-app.com/service/marathon/v2/groups/baz', 'group')


def test_response_error_message_with_other_status_no_json():
    _assert_response_error_message_with_other_status_no_json(
        status_code=401,
        reason=_REASON_1,
        request_url=_URL_1)

    _assert_response_error_message_with_other_status_no_json(
        status_code=401,
        reason=_REASON_2,
        request_url=_URL_2)

    _assert_response_error_message_with_other_status_no_json(
        status_code=403,
        reason=_REASON_X,
        request_url=_URL_X)

    _assert_response_error_message_with_other_status_no_json(
        status_code=404,
        reason=_REASON_X,
        request_url=_URL_X)

    _assert_response_error_message_with_other_status_no_json(
        status_code=422,
        reason=_REASON_X,
        request_url=_URL_X)


def test_response_error_message_with_other_status_json_has_message():
    _assert_response_error_message_with_other_status_json_has_message(
        status_code=401,
        json_message=_MESSAGE_1)

    _assert_response_error_message_with_other_status_json_has_message(
        status_code=401,
        json_message=_MESSAGE_2)

    _assert_response_error_message_with_other_status_json_has_message(
        status_code=403,
        json_message=_MESSAGE_X)

    _assert_response_error_message_with_other_status_json_has_message(
        status_code=404,
        json_message=_MESSAGE_X)

    _assert_response_error_message_with_other_status_json_has_message(
        status_code=422,
        json_message=_MESSAGE_X)


def test_res_err_msg_with_other_status_json_no_message_has_valid_errors():
    _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code=401,
        errors_json=[{'error': 'err1'}, {'error': 'err2'}],
        errors_str='err1\nerr2')

    _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code=401,
        errors_json=[{'error': 'foo'}, {'error': 'bar'}, {'error': 'baz'}],
        errors_str='foo\nbar\nbaz')

    _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code=403,
        errors_json=[{'error': 'foo'}, {'error': 'bar'}, {'error': 'baz'}],
        errors_str='foo\nbar\nbaz')

    _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code=404,
        errors_json=[{'error': 'foo'}, {'error': 'bar'}, {'error': 'baz'}],
        errors_str='foo\nbar\nbaz')

    _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code=422,
        errors_json=[{'error': 'foo'}, {'error': 'bar'}, {'error': 'baz'}],
        errors_str='foo\nbar\nbaz')


def test_res_err_msg_with_other_status_invalid_error_json():
    # Is not an object
    _assert_res_err_msg_with_other_status_invalid_json(401, 'Error!')
    _assert_res_err_msg_with_other_status_invalid_json(401, ['Error!'])
    # Missing both message and errors fields
    _assert_res_err_msg_with_other_status_invalid_json(401, {})
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'foo': 5, 'bar': 'x'})
    # Has non-str message
    _assert_res_err_msg_with_other_status_invalid_json(401, {'message': 42})
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'message': ['Oops!']})
    # Has non-array errors
    _assert_res_err_msg_with_other_status_invalid_json(401, {'errors': 42})
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'errors': {'error': 5}})
    # Errors array has non-object elements
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'errors': [42, True]})
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'errors': [{'error': 'BOOM!'}, 'not_an_error']})
    # Errors array has objects without `error` field
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'errors': [{'error': 'BOOM!'}, {'foo': 'bar'}]})
    # Errors array has objects with non-str `error` field
    _assert_res_err_msg_with_other_status_invalid_json(
        401, {'errors': [{'error': 'BOOM!'}, {'error': 42}]})
    # Other status codes
    _assert_res_err_msg_with_other_status_invalid_json(
        403, {'errors': [{'error': 'BOOM!'}, {'error': 42}]})
    _assert_res_err_msg_with_other_status_invalid_json(
        404, {'errors': [{'error': 'BOOM!'}, {'error': 42}]})
    _assert_res_err_msg_with_other_status_invalid_json(
        422, {'errors': [{'error': 'BOOM!'}, {'error': 42}]})


def _assert_add_pod_puts_json_in_request_body(pod_json):
    rpc_client = mock.create_autospec(rpcclient.RpcClient)

    client = marathon.Client(rpc_client)
    client.add_pod(pod_json)

    rpc_client.http_req.assert_called_with(http.post, 'v2/pods', json=pod_json)


def _assert_add_pod_returns_parsed_response_body(response_json):

    headers = {'Marathon-Deployment-Id': "XYZ"}
    mock_response = _pod_response_fixture(headers)
    mock_response.json.return_value = response_json

    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.return_value = mock_response

    assert marathon_client.add_pod({'some': 'json'}) == "XYZ"


def _assert_remove_pod_builds_rpc_correctly(pod_id, force, path, params):
    marathon_client, rpc_client = _create_fixtures()
    marathon_client.remove_pod(pod_id, force)
    rpc_client.http_req.assert_called_with(http.delete, path, params=params)


def _assert_show_pod_builds_rpc_correctly(pod_id, path):
    marathon_client, rpc_client = _create_fixtures()
    marathon_client.show_pod(pod_id)
    rpc_client.http_req.assert_called_with(http.get, path)


def _assert_show_pod_returns_response_json(expected):
    marathon_client, rpc_client = _create_fixtures()
    mock_response = mock.create_autospec(requests.Response)
    mock_response.json.return_value = expected
    rpc_client.http_req.return_value = mock_response

    response_json = marathon_client.show_pod('arbitrary-id')

    assert response_json == expected


def _assert_list_pod_returns_success_response_json(body_json):
    marathon_client, rpc_client = _create_fixtures()
    mock_response = mock.create_autospec(requests.Response)
    mock_response.json.return_value = body_json
    rpc_client.http_req.return_value = mock_response

    assert marathon_client.list_pod() == body_json


def _assert_update_pod_executes_successfully(
        pod_id, pod_json, force, path, params, deployment_id):
    headers = {'Marathon-Deployment-Id': deployment_id}
    mock_response = _pod_response_fixture(headers)

    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.return_value = mock_response

    actual_return = marathon_client.update_pod(pod_id, pod_json, force)

    rpc_client.http_req.assert_called_with(
        http.put, path, params=params, json=pod_json)
    assert actual_return == deployment_id


def _assert_method_raises_dcos_exception_for_json_parse_errors(invoke_method):
    def assert_test_case(non_json):
        mock_response = mock.create_autospec(requests.Response)
        mock_response.json.side_effect = Exception()
        mock_response.text = non_json

        marathon_client, rpc_client = _create_fixtures()
        rpc_client.http_req.return_value = mock_response

        with pytest.raises(DCOSException) as exception_info:
            invoke_method(marathon_client)

        pattern = ('Error: Response from Marathon was not in expected JSON '
                   'format:\n(.*)')
        actual_error = str(exception_info.value)
        _assert_matches_with_groups(pattern, actual_error, (non_json,))

    assert_test_case('not-json')
    assert_test_case('{"oops"}')


def _assert_add_pod_raises_dcos_exception_if_deployment_id_missing(headers):
    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.return_value = _pod_response_fixture(headers)
    result = marathon_client.add_pod({'some': 'json'})

    assert result is None


def _assert_update_pod_raises_dcos_exception_if_deployment_id_missing(headers):
    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.return_value = _pod_response_fixture(headers)

    with pytest.raises(DCOSException) as exception_info:
        marathon_client.update_pod('foo', {'some': 'json'})

    expected_error = ('Error: missing "Marathon-Deployment-Id" header from '
                      'Marathon response')
    assert str(exception_info.value) == expected_error


def _assert_method_propagates_rpc_dcos_exception(invoke_method):
    marathon_client, rpc_client = _create_fixtures()
    rpc_client.http_req.side_effect = DCOSException('BOOM!')

    with pytest.raises(DCOSException) as exception_info:
        invoke_method(marathon_client)

    assert str(exception_info.value) == 'BOOM!'


@mock.patch('dcos.http.head')
def _assert_pod_feature_supported_raises_exception(head_fn, exception):
    rpc_client = rpcclient.RpcClient('http://base/url', timeout=22)
    marathon_client = marathon.Client(rpc_client)
    head_fn.side_effect = exception

    with pytest.raises(exception.__class__) as exception_info:
        marathon_client.pod_feature_supported()

    assert exception_info.value == exception


def _assert_response_error_message_with_400_status_no_json(
        reason, request_method, request_url):
    message = rpcclient.RpcClient.response_error_message(
        status_code=400,
        reason=reason,
        request_method=request_method,
        request_url=request_url,
        json_body=None)

    pattern = r'Error on request \[(.*) (.*)\]: HTTP 400: (.*)'
    groups = (request_method, request_url, reason)
    _assert_matches_with_groups(pattern, message, groups)


def _assert_response_error_message_with_400_status_json(
        response_json, printed_json):
    message = rpcclient.RpcClient.response_error_message(
        status_code=400,
        reason=_REASON_X,
        request_method=_METHOD_X,
        request_url=_URL_X,
        json_body=response_json)

    error_line, json_lines = message.split('\n', 1)
    pattern = r'Error on request \[(.*) (.*)\]: HTTP 400: (.*):'
    groups = (_METHOD_X, _URL_X, _REASON_X)

    _assert_matches_with_groups(pattern, error_line, groups)
    assert json_lines == printed_json


def _assert_response_error_message_with_other_status_no_json(
        status_code, reason, request_url):
    message = rpcclient.RpcClient.response_error_message(
        status_code=status_code,
        reason=reason,
        request_method=_METHOD_X,
        request_url=request_url,
        json_body=None)

    pattern = r'Error decoding response from \[(.*)\]: HTTP (.*): (.*)'
    groups = (request_url, str(status_code), reason)
    _assert_matches_with_groups(pattern, message, groups)


def _assert_response_error_message_with_other_status_json_has_message(
        status_code, json_message):
    error_message = rpcclient.RpcClient.response_error_message(
        status_code=status_code,
        reason=_REASON_X,
        request_method=_METHOD_X,
        request_url=_URL_X,
        json_body={'message': json_message})

    pattern = r'Error: (.*)'
    _assert_matches_with_groups(pattern, error_message, (json_message,))


def _assert_res_err_msg_with_other_status_json_no_message_has_valid_errors(
        status_code, errors_json, errors_str):
    message = rpcclient.RpcClient.response_error_message(
        status_code=status_code,
        reason=_REASON_X,
        request_method=_METHOD_X,
        request_url=_URL_X,
        json_body={'errors': errors_json})

    pattern = (r'Service likely misconfigured\. Please check your proxy or '
               r'Service URL settings\. See dcos config --help\. (.*)')
    _assert_matches_with_groups(pattern, message, (errors_str,))


def _assert_res_err_msg_with_other_status_invalid_json(status_code, json_body):
    actual = rpcclient.RpcClient.response_error_message(
        status_code=status_code,
        reason=_REASON_X,
        request_method=_METHOD_X,
        request_url=_URL_X,
        json_body=json_body)

    expected = ('Service likely misconfigured. Please check your proxy or '
                'Service URL settings. See dcos config --help. ')
    assert actual == expected


def _assert_matches_with_groups(pattern, text, groups):
    match = re.fullmatch(pattern, text, flags=re.DOTALL)
    assert match
    assert match.groups() == groups


def _create_fixtures():
    rpc_client = mock.create_autospec(rpcclient.RpcClient)
    marathon_client = marathon.Client(rpc_client)

    return marathon_client, rpc_client


def _pod_response_fixture(headers=None):
    mock_response = mock.create_autospec(requests.Response)

    headers = CaseInsensitiveDict({} if headers is None else headers)
    mock_response.headers = headers

    return mock_response


_MESSAGE_1 = 'Oops!'
_MESSAGE_2 = 'Uh oh!'
_MESSAGE_X = "D'oh!"

_METHOD_1 = 'FOO'
_METHOD_2 = 'BAR'
_METHOD_X = 'ANY'

_PRINTED_JSON_1 = (
    '{\n'
    '  "x": "err",\n'
    '  "y": 3.14,\n'
    '  "z": [\n'
    '    1,\n'
    '    2,\n'
    '    3\n'
    '  ]\n'
    '}')
_PRINTED_JSON_2 = (
    '[\n'
    '  "something",\n'
    '  "completely",\n'
    '  "different"\n'
    ']')

_REASON_1 = 'Something Bad'
_REASON_2 = 'Another Reason'
_REASON_X = 'Arbitrary Reason'

_RESPONSE_JSON_1 = {"z": [1, 2, 3], "y": 3.14, "x": "err"}
_RESPONSE_JSON_2 = ["something", "completely", "different"]

_URL_1 = 'http://request/url'
_URL_2 = 'https://another/url'
_URL_X = 'http://does/not/matter'



import os

import docopt

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

import dcoscli

from dcos import cluster, cmds, config, emitting, http, util
from dcos.errors import DCOSAuthenticationException, DCOSException
from dcoscli.auth.main import login
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.tables import clusters_table
from dcoscli.util import confirm, decorate_docopt_usage


emitter = emitting.FlatEmitter()
logger = util.get_logger(__name__)


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("cluster"),
        argv=argv,
        version='dcos-cluster version {}'.format(dcoscli.version))

    http.silence_requests_warnings()

    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: all the supported commands
    :rtype: list of dcos.cmds.Command
    """

    return [
        cmds.Command(
            hierarchy=['cluster', 'setup'],
            arg_keys=['<dcos_url>',
                      '--insecure', '--no-check', '--ca-certs',
                      '--password', '--password-env', '--password-file',
                      '--provider', '--username', '--private-key'],
            function=setup),

        cmds.Command(
            hierarchy=['cluster', 'list'],
            arg_keys=['--json', '--attached'],
            function=_list),

        cmds.Command(
            hierarchy=['cluster', 'remove'],
            arg_keys=['<name>'],
            function=_remove),

        cmds.Command(
            hierarchy=['cluster', 'attach'],
            arg_keys=['<name>'],
            function=_attach),

        cmds.Command(
            hierarchy=['cluster', 'rename'],
            arg_keys=['<name>', '<new_name>'],
            function=_rename),

        cmds.Command(
            hierarchy=['cluster'],
            arg_keys=['--info'],
            function=_info),
    ]


def _info(info):
    """
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: process status
    :rtype: int
    """

    emitter.publish(default_command_info("cluster"))
    return 0


def _list(json_, attached):
    """
    List configured clusters.

    :param json_: output json if True
    :type json_: bool
    :param attached: return only attached cluster
    :type attached: True
    :rtype: None
    """

    clusters = [c.dict() for c in cluster.get_clusters()
                if not attached or c.is_attached()]
    if json_:
        emitter.publish(clusters)
    elif len(clusters) == 0:
        if attached:
            msg = ("No cluster is attached. "
                   "Please run `dcos cluster attach <cluster-name>")
        else:
            msg = ("No clusters are currently configured. "
                   "To configure one, run `dcos cluster setup <dcos_url>`")
        raise DCOSException(msg)
    else:
        emitter.publish(clusters_table(clusters))

    return


def _remove(name):
    """
    :param name: name of cluster
    :type name: str
    :rtype: None
    """

    return cluster.remove(name)


def _attach(name):
    """
    :param name: name of cluster
    :type name: str
    :rtype: None
    """

    c = cluster.get_cluster(name)
    if c is not None:
        return cluster.set_attached(c.get_cluster_path())
    else:
        raise DCOSException("Cluster [{}] does not exist".format(name))


def _rename(name, new_name):
    """
    :param name: name of cluster
    :type name: str
    :param new_name: new_name of cluster
    :type new_name: str
    :rtype: None
    """

    c = cluster.get_cluster(name)
    other = cluster.get_cluster(new_name)
    if c is None:
        raise DCOSException("Cluster [{}] does not exist".format(name))
    elif other and other != c:
        msg = "A cluster with name [{}] already exists"
        raise DCOSException(msg.format(new_name))
    else:
        config.set_val("cluster.name", new_name, c.get_config_path())


def setup(dcos_url,
          insecure=False, no_check=False, ca_certs=None,
          password_str=None, password_env=None, password_file=None,
          provider=None, username=None, key_path=None):
    """
    Setup the CLI to talk to your DC/OS cluster.

    :param dcos_url: master ip of cluster
    :type dcos_url: str
    :param insecure: whether or not to verify ssl certs
    :type insecure: bool
    :param no_check: whether or not to verify downloaded ca cert
    :type no_check: bool
    :param ca_certs: path to root CA to verify requests
    :type ca_certs: str
    :param password_str: password
    :type password_str: str
    :param password_env: name of environment variable with password
    :type password_env: str
    :param password_file: path to file with password
    :type password_file: bool
    :param provider: name of provider to authentication with
    :type provider: str
    :param username: username
    :type username: str
    :param key_path: path to file with private key
    :type param: str
    :returns: process status
    :rtype: int
    """

    with cluster.setup_directory() as temp_path:

        # set cluster as attached
        cluster.set_attached(temp_path)

        # authenticate
        config.set_val("core.dcos_url", dcos_url)
        # get validated dcos_url
        dcos_url = config.get_config_val("core.dcos_url")

        # configure ssl settings
        stored_cert = False
        if insecure:
            config.set_val("core.ssl_verify", "false")
        elif ca_certs:
            config.set_val("core.ssl_verify", ca_certs)
        else:
            cert = cluster.get_cluster_cert(dcos_url)
            # if we don't have a cert don't try to verify one
            if cert is False:
                config.set_val("core.ssl_verify", "false")
            else:
                stored_cert = _store_cluster_cert(cert, no_check)

        try:
            login(dcos_url,
                  password_str, password_env, password_file,
                  provider, username, key_path)
        except DCOSAuthenticationException:
            msg = ("Authentication failed. "
                   "Please run `dcos cluster setup <dcos_url>`")
            raise DCOSException(msg)

        # configure cluster directory
        cluster.setup_cluster_config(dcos_url, temp_path, stored_cert)

    return 0


def _user_cert_validation(cert_str):
    """Prompt user for validation of certification from cluster

    :param cert_str: cluster certificate bundle
    :type cert_str: str
    :returns whether or not user validated cert
    :rtype: bool
    """

    cert = x509.load_pem_x509_certificate(
        cert_str.encode('utf-8'), default_backend())
    fingerprint = cert.fingerprint(hashes.SHA256())
    pp_fingerprint = ":".join("{:02x}".format(c) for c in fingerprint).upper()

    msg = "SHA256 fingerprint of cluster certificate bundle:\n{}".format(
            pp_fingerprint)

    return confirm(msg, False)


def _store_cluster_cert(cert, no_check):
    """Store cluster certificate bundle downloaded from cluster and store
    settings in core.ssl_verify

    :param cert: ca cert from cluster
    :type cert: str
    :param no_check: whether to verify downloaded cert
    :type no_check: bool
    :returns: whether or not we are storing the downloaded cert bundle
    :rtype: bool
    """

    if not no_check:
        if not _user_cert_validation(cert):
            # we don't have a cert, but we still want to validate SSL
            config.set_val("core.ssl_verify", "true")
            return False

    with util.temptext() as temp_file:
        _, temp_path = temp_file

        with open(temp_path, 'w') as f:
            f.write(cert)

        cert_path = os.path.join(
            config.get_attached_cluster_path(), "dcos_ca.crt")

        util.sh_copy(temp_path, cert_path)

    config.set_val("core.ssl_verify", cert_path)
    return True



import os
import signal
import sys

import docopt
from six.moves import urllib

import dcoscli
from dcos import (cluster, config, constants, emitting, errors, http,
                  subcommand, util)
from dcos.errors import DCOSException
from dcoscli.help.main import dcos_help
from dcoscli.subcommand import default_doc, SubcommandMain

logger = util.get_logger(__name__)
emitter = emitting.FlatEmitter()


def main():
    try:
        return _main()
    except DCOSException as e:
        emitter.publish(e)
        return 1


def _get_versions():
    """Print DC/OS and DC/OS-CLI versions

    :returns: Process status
    :rtype: int
    """

    dcos_info = {}
    try:
        dcos_url = config.get_config_val("core.dcos_url")
        url = urllib.parse.urljoin(
            dcos_url, 'dcos-metadata/dcos-version.json')
        res = http.get(url, timeout=1)
        if res.status_code == 200:
            dcos_info = res.json()
    except Exception as e:
        logger.exception(e)
        pass

    emitter.publish(
        "dcoscli.version={}\n".format(dcoscli.version) +
        "dcos.version={}\n".format(dcos_info.get("version", "N/A")) +
        "dcos.commit={}\n".format(dcos_info.get(
            "dcos-image-commit", "N/A")) +
        "dcos.bootstrap-id={}".format(dcos_info.get("bootstrap-id", "N/A"))
    )
    return 0


def _main():
    signal.signal(signal.SIGINT, signal_handler)

    http.silence_requests_warnings()

    args = docopt.docopt(default_doc("dcos"), options_first=True)

    log_level = args['--log-level']
    if log_level and not _config_log_level_environ(log_level):
        return 1

    if args['--debug']:
        os.environ[constants.DCOS_DEBUG_ENV] = 'true'

    util.configure_process_from_environ()

    if config.uses_deprecated_config():
        cluster.move_to_cluster_config()

    if args['--version']:
        return _get_versions()

    command = args['<command>']

    if not command:
        if args['--help']:
            command = "help"
        else:
            return dcos_help()

    if command in subcommand.default_subcommands():
        sc = SubcommandMain(command, args['<args>'])
    else:
        executable = subcommand.command_executables(command)
        sc = subcommand.SubcommandProcess(
            executable, command, args['<args>'])

    exitcode, _ = sc.run_and_capture()
    return exitcode


def _config_log_level_environ(log_level):
    """
    :param log_level: Log level to set
    :type log_level: str
    :returns: True if the log level was configured correctly; False otherwise.
    :rtype: bool
    """

    log_level = log_level.lower()

    if log_level in constants.VALID_LOG_LEVEL_VALUES:
        os.environ[constants.DCOS_LOG_LEVEL_ENV] = log_level
        return True

    msg = 'Log level set to an unknown value {!r}. Valid values are {!r}'
    emitter.publish(msg.format(log_level, constants.VALID_LOG_LEVEL_VALUES))

    return False


def signal_handler(signal, frame):
    emitter.publish(
        errors.DefaultError("User interrupted command with Ctrl-C"))
    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())




def auth_provider_fixture():
    """Auth provider fixture

    :rtype: dict
    """

    return {
        "dcos-users": {
            "authentication-type": "dcos-uid-password",
            "client-method": "dcos-usercredential-post-receive-authtoken",
            "config": {
                "start_flow_url": "/acs/api/v1/auth/login"
            },
            "description": "Default DC/OS authenticator"
        }
    }



import contextlib
import json
import re

from .common import assert_command, exec_command


def add_app(app_path, wait=True):
    """ Add an app, and wait for it to deploy

    :param app_path: path to app's json definition
    :type app_path: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    cmd = ['dcos', 'marathon', 'app', 'add', app_path]
    returncode, stdout, stderr = exec_command(cmd)
    assert returncode == 0
    assert re.fullmatch('Created deployment \S+\n', stdout.decode('utf-8'))
    assert stderr == b''

    if wait:
        watch_all_deployments()


def start_app(app_id, instances=None):
    cmd = ['dcos', 'marathon', 'app', 'start', app_id]
    if instances is not None:
        cmd.append(str(instances))

    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 0
    assert stdout.decode().startswith('Created deployment ')
    assert stderr == b''


def list_apps(app_id=None):
    returncode, stdout, stderr = exec_command(
        ['dcos', 'marathon', 'app', 'list', '--json'])

    result = json.loads(stdout.decode('utf-8'))

    if app_id is None:
        assert len(result) == 0
    else:
        assert len(result) == 1
        assert result[0]['id'] == '/' + app_id

    assert returncode == 0
    assert stderr == b''

    return result


def remove_group(group_id):
    assert_command(['dcos', 'marathon', 'group', 'remove', group_id])

    # Let's make sure that we don't return until the deployment has finished
    watch_all_deployments()


def remove_app(app_id):
    """ Remove an app

    :param app_id: id of app to remove
    :type app_id: str
    :rtype: None
    """

    assert_command(['dcos', 'marathon', 'app', 'remove', '--force', app_id])


def show_group(group_id, version=None):
    if version is None:
        cmd = ['dcos', 'marathon', 'group', 'show', group_id]
    else:
        cmd = ['dcos', 'marathon', 'group', 'show',
               '--group-version={}'.format(version), group_id]

    returncode, stdout, stderr = exec_command(cmd)

    result = json.loads(stdout.decode('utf-8'))

    assert returncode == 0
    assert isinstance(result, dict)
    assert result['id'] == '/' + group_id
    assert stderr == b''

    return result


def remove_pod(pod_id, force=True):
    """ Remove a pod

    :param pod_id: id of app to remove
    :type pod_id: str
    :param force: whether to force a remove
    :type force: bool
    :rtype: None
    """

    cmd = ['dcos', 'marathon', 'pod', 'remove', pod_id]
    if force:
        cmd += ['--force']
    assert_command(cmd)


def show_app(app_id, version=None):
    """Show details of a Marathon application.

    :param app_id: The id for the application
    :type app_id: str
    :param version: The version, either absolute (date-time) or relative
    :type version: str
    :returns: The requested Marathon application
    :rtype: dict
    """

    if version is None:
        cmd = ['dcos', 'marathon', 'app', 'show', app_id]
    else:
        cmd = ['dcos', 'marathon', 'app', 'show',
               '--app-version={}'.format(version), app_id]

    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 0
    assert stderr == b''

    result = json.loads(stdout.decode('utf-8'))
    assert isinstance(result, dict)
    assert result['id'] == '/' + app_id

    return result


@contextlib.contextmanager
def app(path, app_id, wait=True):
    """Context manager that deploys an app on entrance, and removes it on
    exit.

    :param path: path to app's json definition:
    :type path: str
    :param app_id: app id
    :type app_id: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    add_app(path, wait)
    try:
        yield
    finally:
        remove_app(app_id)
        watch_all_deployments()


def add_pod(pod_path, wait=True):
    """Add a pod, and wait for it to deploy

    :param pod_path: path to pod's json definition
    :type pod_path: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    cmd = ['dcos', 'marathon', 'pod', 'add', pod_path]
    returncode, stdout, stderr = exec_command(cmd)
    assert returncode == 0
    assert re.fullmatch('Created deployment \S+\n', stdout.decode('utf-8'))
    assert stderr == b''

    if wait:
        watch_all_deployments()


def pod_spec_json(expected_pod_spec, actual_pod_spec):
    """Checks that the "actual" pod spec JSON matches the "expected" JSON.

    The comparison only looks at specific fields that are present in the
    test data used by this module.

    :param expected_pod_spec: contains the baseline values for the comparison
    :type expected_pod_spec: {}
    :param actual_pod_spec: has its fields checked against the expected fields
    :type actual_pod_spec: {}
    :rtype: None
    """

    expected_containers = expected_pod_spec['containers']
    actual_containers = actual_pod_spec['containers']
    actual_containers_by_name = {c['name']: c for c in actual_containers}

    for expected_container in expected_containers:
        container_name = expected_container['name']
        actual_container = actual_containers_by_name[container_name]

        for k, v in expected_container['resources'].items():
            assert actual_container['resources'][k] == v

    assert len(actual_containers) == len(expected_containers)


def pod_status_json(expected_pod_status, actual_pod_status):
    """Checks that the "actual" pod status JSON matched the "expected" JSON.

    The comparison only looks at specific fields that are present in the
    test data used by this module.

    :param expected_pod_status: contains the baseline values for the comparison
    :type expected_pod_status: {}
    :param actual_pod_status: has its fields checked against expected's fields
    :type actual_pod_status: {}
    :rtype: None
    """

    assert actual_pod_status['id'] == expected_pod_status['id']
    assert actual_pod_status['status'] == expected_pod_status['status']
    assert len(actual_pod_status['instances']) == \
        len(expected_pod_status['instances'])

    pod_spec_json(expected_pod_status['spec'],
                  actual_pod_status['spec'])

    expected_instance = expected_pod_status['instances'][0]
    expected_container_statuses = {container['name']: container['status']
                                   for container
                                   in expected_instance['containers']}

    for actual_instance in actual_pod_status['instances']:
        assert actual_instance['status'] == expected_instance['status']

        actual_container_statuses = {container['name']: container['status']
                                     for container
                                     in actual_instance['containers']}

        assert actual_container_statuses == expected_container_statuses


def show_pod(pod_id, expected_json):
    """Show details of a Marathon pod and make sure it matches expected output

    :param pod_id: The id for the pod
    :type pod_id: str
    :param expected_json: expected results for pod `show`
    :type expected_json: dict
    :rtype: None
    """

    cmd = ['dcos', 'marathon', 'pod', 'show', pod_id]
    returncode, stdout, stderr = exec_command(cmd)

    assert returncode == 0
    assert stderr == b''

    status_json = json.loads(stdout.decode('utf-8'))
    pod_status_json(expected_json, status_json)


def add_group(group_path, wait=True):
    """Add a group, and wait for it to deploy

    :param group_path: path to pod's json definition
    :type group_path: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    cmd = ['dcos', 'marathon', 'group', 'add', group_path]
    returncode, stdout, stderr = exec_command(cmd)
    assert returncode == 0
    assert re.fullmatch('Created deployment \S+\n', stdout.decode('utf-8'))
    assert stderr == b''

    if wait:
        watch_all_deployments()


@contextlib.contextmanager
def group(path, group_id, wait=True):
    """Context manager that deploys an group on entrance, and removes it on exit

    :param path: path to group's json definition:
    :type path: str
    :param group_id: group id
    :type group_id: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    add_group(path, wait)
    try:
        yield
    finally:
        remove_group(group_id)
        watch_all_deployments()


@contextlib.contextmanager
def pod(path, pod_id, wait=True):
    """Context manager that deploys an pod on entrance, and removes it on exit

    :param path: path to pod's json definition:
    :type path: str
    :param pod_id: pod id
    :type pod_id: str
    :param wait: whether to wait for the deploy
    :type wait: bool
    :rtype: None
    """

    add_pod(path, wait)
    try:
        yield
    finally:
        remove_pod(pod_id)
        watch_all_deployments()


@contextlib.contextmanager
def pods(pods):
    """Context manager that deploys pods on entrance, and removes
    them on exit.

    :param pods: dict of path/to/pod/json -> pod id
    :type pods: {}
    :rtype: None
    """

    for pod_path in pods:
        add_pod(pod_path, wait=False)
    watch_all_deployments()

    try:
        yield
    finally:
        for pod_id in list(pods.values()):
            remove_pod(pod_id)
        watch_all_deployments()


def watch_deployment(deployment_id, count):
    """Wait for a deployment to complete.

    :param deployment_id: deployment id
    :type deployment_id: str
    :param count: max number of seconds to wait
    :type count: int
    :rtype: None
    """

    returncode, stdout, stderr = exec_command(
        ['dcos', 'marathon', 'deployment', 'watch',
            '--max-count={}'.format(count), deployment_id])

    assert returncode == 0
    assert stderr == b''


def watch_all_deployments(count=300):
    """Wait for all deployments to complete.

    :param count: max number of seconds to wait
    :type count: int
    :rtype: None
    """

    deps = list_deployments()
    for dep in deps:
        watch_deployment(dep['id'], count)


def list_deployments(expected_count=None, app_id=None):
    """Get all active deployments.

    :param expected_count: assert that number of active deployments
    equals `expected_count`
    :type expected_count: int
    :param app_id: only get deployments for this app
    :type app_id: str
    :returns: active deployments
    :rtype: [dict]
    """

    cmd = ['dcos', 'marathon', 'deployment', 'list', '--json']
    if app_id is not None:
        cmd.append(app_id)

    returncode, stdout, stderr = exec_command(cmd)

    result = json.loads(stdout.decode('utf-8'))

    assert returncode == 0
    if expected_count is not None:
        assert len(result) == expected_count
    assert stderr == b''

    return result




def job_list_fixture():
    """Job list fixture

    :rtype: dict
    """

    return [
      {
        "description": "sleeping is what is do",
        "historySummary": {
          "failureCount": 0,
          "lastFailureAt": None,
          "lastSuccessAt": None,
          "successCount": 0
        },
        "id": "snorlax",
        "labels": {},
        "run": {
          "artifacts": [],
          "cmd": "sleep 10",
          "cpus": 0.01,
          "disk": 0,
          "env": {},
          "maxLaunchDelay": 3600,
          "mem": 32,
          "placement": {
            "constraints": []
          },
          "restart": {
            "policy": "NEVER"
          },
          "volumes": []
        }
      },
      {
        "description": "electrifying rodent",
        "historySummary": {
          "failureCount": 0,
          "lastFailureAt": None,
          "lastSuccessAt": "2017-03-31T14:22:01.541+0000",
          "successCount": 1
        },
        "id": "pikachu",
        "labels": {},
        "run": {
          "artifacts": [],
          "cmd": "sleep 10",
          "cpus": 0.01,
          "disk": 0,
          "env": {},
          "maxLaunchDelay": 3600,
          "mem": 32,
          "placement": {
            "constraints": []
          },
          "restart": {
            "policy": "NEVER"
          },
          "volumes": []
        }
      }]


def job_run_fixture():
    """Job run fixture

    :rtype: dict
    """

    return [
      {
        "completedAt": None,
        "createdAt": "2017-03-31T21:05:30.613+0000",
        "id": "20170331210530QHpRU",
        "jobId": "pikachu",
        "status": "ACTIVE",
        "tasks": [
          {
            "id": "pikachu_20170331210530QHpRU.c5e4b1e7-1655-11e7-8bd5-6ef119b8e20f",  # NOQA
            "startedAt": "2017-03-31T21:05:31.499+0000",
            "status": "TASK_RUNNING"
          }
        ]
      },
      {
        "completedAt": None,
        "createdAt": "2017-03-31T21:05:32.422+0000",
        "id": "20170331210532uxgVF",
        "jobId": "pikachu",
        "status": "ACTIVE",
        "tasks": [
          {
            "id": "pikachu_20170331210532uxgVF.c8e324d8-1655-11e7-8bd5-6ef119b8e20f",  # NOQA
            "startedAt": "2017-03-31T21:05:36.417+0000",
            "status": "TASK_RUNNING"
          }
        ]
      }]


def job_history_fixture():
    """Job history fixture

    :rtype: dict
    """

    return [
      {
        "createdAt": "2017-03-31T21:05:32.422+0000",
        "finishedAt": "2017-03-31T21:05:46.805+0000",
        "id": "20170331210532uxgVF"
      },
      {
        "createdAt": "2017-03-31T21:05:30.613+0000",
        "finishedAt": "2017-03-31T21:05:41.740+0000",
        "id": "20170331210530QHpRU"
      }]


def job_schedule_fixture():
    """Job schedule fixture

    :rtype: dict
    """

    return [
      {
        "concurrencyPolicy": "ALLOW",
        "cron": "20 0 * * *",
        "enabled": True,
        "id": "nightly",
        "nextRunAt": "2017-04-01T00:20:00.000+0000",
        "startingDeadlineSeconds": 900,
        "timezone": "UTC"
      }]



import json
import os
import sys

import docopt
import pkg_resources
import six
from six.moves import urllib

import dcoscli
from dcos import (cmds, config, emitting, http,
                  metronome, options, packagemanager, util)
from dcos.cosmos import get_cosmos_url
from dcos.errors import DCOSException, DCOSHTTPException
from dcoscli import tables
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import decorate_docopt_usage


logger = util.get_logger(__name__)
emitter = emitting.FlatEmitter()

DEFAULT_TIMEOUT = 180

# single job, not a lot of data
EMBEDS_FOR_JOB_HISTORY = [
    metronome.EMBED_ACTIVE_RUNS,
    metronome.EMBED_SCHEDULES,
    metronome.EMBED_HISTORY]

# unknown number of jobs, using history summary
EMBEDS_FOR_JOBS_HISTORY = [
    metronome.EMBED_ACTIVE_RUNS,
    metronome.EMBED_HISTORY_SUMMARY]


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("job"),
        argv=argv,
        version='dcos-job version {}'.format(dcoscli.version))

    return cmds.execute(_cmds(), args)


def _check_capability():
    """
    The function checks if cluster has metronome capability.

    :raises: DCOSException if cluster does not have metronome capability
    """

    cosmos = packagemanager.PackageManager(get_cosmos_url())
    if not cosmos.has_capability('METRONOME'):
        raise DCOSException(
            'DC/OS backend does not support metronome capabilities in this '
            'version. Must be DC/OS >= 1.8')


def _cmds():
    """
    :returns: all the supported commands
    :rtype: dcos.cmds.Command
    """

    return [

        cmds.Command(
            hierarchy=['job', 'run'],
            arg_keys=['<job-id>', '--json'],
            function=_run),

        cmds.Command(
            hierarchy=['job', 'kill'],
            arg_keys=['<job-id>', '<run-id>', '--all'],
            function=_kill),

        cmds.Command(
            hierarchy=['job', 'schedule', 'add'],
            arg_keys=['<job-id>', '<schedule-file>'],
            function=_add_schedule),

        cmds.Command(
            hierarchy=['job', 'schedule', 'update'],
            arg_keys=['<job-id>', '<schedule-file>'],
            function=_update_schedules),

        cmds.Command(
            hierarchy=['job', 'schedule', 'show'],
            arg_keys=['<job-id>', '--json'],
            function=_show_schedule),

        cmds.Command(
            hierarchy=['job', 'show', 'runs'],
            arg_keys=['<job-id>', '<run-id>', '--json', '--q'],
            function=_show_runs),

        cmds.Command(
            hierarchy=['job', 'schedule', 'remove'],
            arg_keys=['<job-id>', '<schedule-id>'],
            function=_remove_schedule),

        cmds.Command(
            hierarchy=['job', 'list'],
            arg_keys=['--json'],
            function=_list),

        cmds.Command(
            hierarchy=['job', 'history'],
            arg_keys=['<job-id>', '--json', '--show-failures'],
            function=_history),

        cmds.Command(
            hierarchy=['job', 'remove'],
            arg_keys=['<job-id>', '--stop-current-job-runs'],
            function=_remove),

        cmds.Command(
            hierarchy=['job', 'add'],
            arg_keys=['<job-file>'],
            function=_add_job),

        cmds.Command(
            hierarchy=['job', 'update'],
            arg_keys=['<job-file>'],
            function=_update_job),

        cmds.Command(
            hierarchy=['job', 'show'],
            arg_keys=['<job-id>'],
            function=_show),

        cmds.Command(
            hierarchy=['job'],
            arg_keys=['--config-schema', '--info'],
            function=_job)
    ]


def _job(config_schema=False, info=False):
    """
    :param config_schema: Whether to output the config schema
    :type config_schema: boolean
    :param info: Whether to output a description of this subcommand
    :type info: boolean
    :returns: process return code
    :rtype: int
    """

    if config_schema:
        schema = _cli_config_schema()
        emitter.publish(schema)
    elif info:
        _info()
    else:
        doc = default_command_info("job")
        raise DCOSException(options.make_generic_usage_message(doc))

    return 0


def _remove_schedule(job_id, schedule_id):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param schedule_id: Id of the schedule
    :type schedule_id: str
    :returns: process return code
    :rtype: int
    """

    try:
        client = metronome.create_client()
        client.remove_schedule(job_id, schedule_id)
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            raise DCOSException("Schedule or job ID does NOT exist.")
    except DCOSException as e:
        raise DCOSException("Unable to remove schedule ID '{}' for job ID '{}'"
                            .format(schedule_id, job_id))

    return 0


def _remove(job_id, stop_current_job_runs=False):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param stop_current_job_runs: If job runs should be stop as
    part of the remove
    :type stop_current_job_runs: boolean
    :returns: process return code
    :rtype: int
    """

    try:
        client = metronome.create_client()
        client.remove_job(job_id, stop_current_job_runs)

    except DCOSHTTPException as e:
        if e.response.status_code == 500 and stop_current_job_runs:
            return _remove(job_id, False)
        else:
            raise DCOSException("Unable to remove '{}'.  It may be running."
                                .format(job_id))
    except DCOSException as e:
        raise DCOSException("Unable to remove '{}'.  It may be running."
                            .format(job_id))

    return 0


def _kill(job_id, run_id, all=False):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: process return code
    :rtype: int
    """
    deadpool = []
    if run_id is None and all is True:
        deadpool = _get_ids(_get_runs(job_id))
    else:
        deadpool.append(run_id)

    client = metronome.create_client()
    for dead in deadpool:
        try:
            client.kill_run(job_id, dead)
        except DCOSHTTPException as e:
            if e.response.status_code == 404:
                raise DCOSException("Job ID or Run ID does NOT exist.")
        except DCOSException as e:
            raise DCOSException("Unable stop run ID '{}' for job ID '{}'"
                                .format(dead, job_id))
        else:
            emitter.publish("Run '{}' for job '{}' killed."
                            .format(dead, job_id))
    return 0


def _list(json_flag=False):
    """ Provides a list of jobs along with their active runs and history summary
    :returns: process return code
    :rtype: int
    """
    client = metronome.create_client()
    json_list = client.get_jobs(EMBEDS_FOR_JOBS_HISTORY)

    if json_flag:
        emitter.publish(json_list)
    else:
        table = tables.job_table(json_list)
        output = six.text_type(table)
        if output:
            emitter.publish(output)

    return 0


def _history(job_id, json_flag=False, show_failures=False):
    """
    :returns: process return code
    :rtype: int
    """

    client = metronome.create_client()
    json_history = client.get_job(job_id, EMBEDS_FOR_JOB_HISTORY)

    if 'history' not in json_history:
        return 0

    if json_flag:
        emitter.publish(json_history)
    else:
        emitter.publish(_get_history_message(json_history, job_id))
        table = tables.job_history_table(
            json_history['history']['successfulFinishedRuns'])
        output = six.text_type(table)
        if output:
            emitter.publish(output)

        if show_failures:
            emitter.publish(_get_history_message(
                            json_history, job_id, False))
            table = tables.job_history_table(
                json_history['history']['failedFinishedRuns'])
            output = six.text_type(table)
            if output:
                emitter.publish(output)

    return 0


def _get_history_message(json_history, job_id, success=True):
    """
    :param json_history: json of history
    :type json_history: json
    :param job_id: Id of the job
    :type job_id: str
    :returns: history message
    :rtype: str
    """
    if success is True:
        return "'{}'  Successful runs: {} Last Success: {}".format(
                job_id, json_history['history']['successCount'],
                json_history['history']['lastSuccessAt'])
    else:
        return "'{}'  Failure runs: {} Last Failure: {}".format(
                job_id, json_history['history']['failureCount'],
                json_history['history']['lastFailureAt'])


def _show(job_id):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: process return code
    :rtype: int
    """

    try:
        client = metronome.create_client()
        json_job = client.get_job(job_id)
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            raise DCOSException("Job ID: '{}' does NOT exist.".format(job_id))
        else:
            raise DCOSException(e)

    emitter.publish(json_job)

    return 0


def _show_runs(job_id, run_id=None, json_flag=False, q=False):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: process return code
    :rtype: int
    """

    json_runs = _get_runs(job_id, run_id)
    if q is True:
        ids = _get_ids(json_runs)
        emitter.publish(ids)
    elif json_flag is True:
        emitter.publish(json_runs)
    else:
        if json_flag:
            emitter.publish(json_runs)
        else:
            if _json_array_has_element(json_runs, 'id'):
                table = tables.job_runs_table(json_runs)
                output = six.text_type(table)
                if output:
                    emitter.publish(output)
            else:
                emitter.publish("Nothing running for '{}'".format(job_id))

    return 0


def _json_array_has_element(json_object, field):
    exists = False
    for element in json_object:
        if field in element:
            exists = True
            break
    return exists


def _get_runs(job_id, run_id=None):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: json of all running instance of a job_id
    :rtype: json
    """

    client = metronome.create_client()
    try:
        if run_id is None:
            return client.get_runs(job_id)
        else:
            return client.get_run(job_id, run_id)
    except DCOSException as e:
        raise DCOSException(e)


def _run(job_id, json_flag=False):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: process return code
    :rtype: int
    """

    try:
        client = metronome.create_client()
        run_job = client.run_job(job_id)
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            emitter.publish("Job ID: '{}' does not exist.".format(job_id))
        else:
            emitter.publish("Error running job: '{}'".format(job_id))
    if json_flag:
        emitter.publish(run_job)
    else:
        emitter.publish('Run ID: {}'.format(run_job['id']))

    return 0


def _show_schedule(job_id, json_flag=False):
    """
    :param job_id: Id of the job
    :type job_id: str
    :returns: process return code
    :rtype: int
    """

    try:
        client = metronome.create_client()
        json_schedule = client.get_schedules(job_id)
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            raise DCOSException("Job ID: '{}' does NOT exist.".format(job_id))
        else:
            raise DCOSException(e)
    except DCOSException as e:
        raise DCOSException(e)

    if json_flag:
        emitter.publish(json_schedule)
    else:
        table = tables.schedule_table(json_schedule)
        output = six.text_type(table)
        if output:
            emitter.publish(output)

    return 0


def parse_schedule_json(schedules_json):
    """
    The original design of metronome had an array of schedules defined but
    limited it to 1.  This limits to 1 and takes the array format or just
    1 schedule format.
    :param schedules_json: schedule or array of schedules in json
    :type schedules_json: json [] or {}
    :returns: schedule json
    :rtype: json
    """
    if type(schedules_json) is list:
        return schedules_json[0]
    else:
        return schedules_json


def _add_schedules(job_id, schedules_json):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param schedules_json: json for the schedules
    :type schedules_json: json
    :returns: process return code
    :rtype: int
    """

    if schedules_json is None:
        raise DCOSException('Schedule JSON is required.')

    schedule = parse_schedule_json(schedules_json)
    client = metronome.create_client()
    client.add_schedule(job_id, schedule)

    return 0


def _update_schedules(job_id, schedules_file):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param schedule_id: Id of the schedule
    :type schedule_id: str
    :param schedule_file: filename for the schedule resource
    :type schedule_file: str
    :returns: process return code
    :rtype: int
    """
    schedules = _get_resource(schedules_file)
    schedule = schedules[0]  # 1 update
    schedule_id = schedule['id']

    return _update_schedule(job_id, schedule_id, schedule)


def _update_schedule(job_id, schedule_id, schedule_json):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param schedule_id: Id of the schedule
    :type schedule_id: str
    :param schedules_json: json for the schedules
    :type schedules_json: json
    :returns: process return code
    :rtype: int
    """

    if schedule_json is None:
        raise DCOSException("No schedule to update.")

    try:
        client = metronome.create_client()
        client.update_schedule(job_id, schedule_id, schedule_json)
        emitter.publish("Schedule ID `{}` for job ID `{}` updated."
                        .format(schedule_id, job_id))
    except DCOSHTTPException as e:
        if e.response.status_code == 404:
            emitter.publish("Job ID: '{}' or schedule ID '{}' does NOT exist."
                            .format(job_id, schedule_id))
    except DCOSException as e:
        raise DCOSException(e)

    return 0


def _add_schedule(job_id, schedule_file):
    """
    :param job_id: Id of the job
    :type job_id: str
    :param schedule_file: filename for the schedule resource
    :type schedule_file: str
    :returns: process return code
    :rtype: int
    """

    schedules = _get_resource(schedule_file)
    return _add_schedules(job_id, schedules)


def _add_job(job_file):
    """
    :param job_file: optional filename for the application resource
    :type job_file: str
    :returns: process return code
    :rtype: int
    """

    full_json = _get_resource(job_file)
    if full_json is None:
        raise DCOSException("No JSON provided.")

    if 'id' not in full_json:
        raise DCOSException("Jobs JSON requires an ID.")

    if 'disk' not in full_json['run']:
        full_json['run']['disk'] = 0

    job_id = full_json['id']
    schedules = None

    if 'schedules' in full_json:
        schedules = full_json['schedules']
        del full_json['schedules']

    client = metronome.create_client()
    client.add_job(full_json)

    if schedules is not None:
        return _add_schedules(job_id, schedules)

    return 0


def _update_job(job_file):
    """
    :param job_file: filename for the application resource
    :type job_file: str
    :returns: process return code
    :rtype: int
    """
    # only updates the job (does NOT update schedules)
    full_json = _get_resource(job_file)
    if full_json is None:
        raise DCOSException("No JSON provided.")

    job_id = full_json['id']

    if 'schedules' in full_json:
        del full_json['schedules']

    try:
        client = metronome.create_client()
        client.update_job(job_id, full_json)
    except DCOSHTTPException as e:
        emitter.publish("Error updating job: '{}'".format(job_id))

    return 0


def _info():
    """
    :returns: process return code
    :rtype: int
    """

    emitter.publish(default_command_info("job"))
    return 0


def _cli_config_schema():
    """
    :returns: schema for metronome cli config
    :rtype: dict
    """
    return json.loads(
        pkg_resources.resource_string(
            'dcoscli',
            'data/config-schema/job.json').decode('utf-8'))


def _do_request(url, method, timeout=None, stream=False, **kwargs):
    """
    make HTTP request

    :param url: url
    :type url: string
    :param method: HTTP method, GET or POST
    :type  method: string
    :param timeout: HTTP request timeout, default 3 seconds
    :type  timeout: integer
    :param stream: stream parameter for requests lib
    :type  stream: bool
    :return: http response
    :rtype: requests.Response
    """

    def _is_success(status_code):
        # consider 400 and 503 to be successful status codes.
        # API will return the error message.
        if status_code in [200, 400, 503]:
            return True
        return False

    if timeout is None:
        timeout = _get_timeout()

    url = urllib.parse.urljoin(_get_metronome_url(), url)
    if method.lower() == 'get':
        http_response = http.get(url, is_success=_is_success,
                                 timeout=timeout, **kwargs)
    elif method.lower() == 'post':
        http_response = http.post(url, is_success=_is_success,
                                  timeout=timeout, stream=stream, **kwargs)
    elif method.lower() == 'delete':
        http_response = http.delete(url, is_success=_is_success,
                                    timeout=timeout, stream=stream, **kwargs)
    else:
        raise DCOSException('Unsupported HTTP method: ' + method)
    return http_response


def _read_http_response_body(http_response):
    """
    Get an requests HTTP response, read it and deserialize to json.

    :param http_response: http response
    :type http_response: requests.Response onject
    :return: deserialized json
    :rtype: dict
    """

    data = b''
    try:
        for chunk in http_response.iter_content(1024):
            data += chunk
        bundle_response = util.load_jsons(data.decode('utf-8'))
        return bundle_response
    except DCOSException:
        raise


def _get_ids(ids_json):
    """
    :param ids_json: json array of elements with ids
    :type ids_json: json
    :returns: set of ids
    :rtype: set
    """
    ids = list()
    for element in ids_json:
        ids.append(element['id'])

    return ids


def _get_resource(resource):
    """
    :param resource: optional filename or http(s) url
    for the application or group resource
    :type resource: str
    :returns: resource
    :rtype: dict
    """
    if resource is not None:
        if os.path.isfile(resource):
            with util.open_file(resource) as resource_file:
                return util.load_json(resource_file)
        else:
            try:
                http.silence_requests_warnings()
                req = http.get(resource)
                if req.status_code == 200:
                    data = b''
                    for chunk in req.iter_content(1024):
                        data += chunk
                    return util.load_jsons(data.decode('utf-8'))
                else:
                    raise DCOSHTTPException("HTTP error code: {}"
                                            .format(req.status_code))
            except Exception:
                logger.exception('Cannot read from resource %s', resource)
                raise DCOSException(
                    "Can't read from resource: {0}.\n"
                    "Please check that it exists.".format(resource))

    # Check that stdin is not tty
    if sys.stdin.isatty():
        # We don't support TTY right now. In the future we will start an
        # editor
        raise DCOSException(
            "We currently don't support reading from the TTY. Please "
            "specify an application JSON.\n"
            "E.g.: dcos job add < app_resource.json")

    return util.load_json(sys.stdin)


def _get_metronome_url(toml_config=None):
    """
    :param toml_config: configuration dictionary
    :type toml_config: config.Toml
    :returns: metronome base url
    :rtype: str
    """
    if toml_config is None:
        toml_config = config.get_config()

    metronome_url = config.get_config_val('metronome.url', toml_config)
    if metronome_url is None:
        # dcos must be capable to use dcos_url
        _check_capability()
        dcos_url = config.get_config_val('core.dcos_url', toml_config)
        if dcos_url is None:
            raise config.missing_config_exception(['core.dcos_url'])
        metronome_url = urllib.parse.urljoin(dcos_url, 'service/metronome/')

    return metronome_url


def _get_api_url(path):
    """
    :param path: service path
    :type path: str
    :returns: metronome base url
    :rtype: str
    """
    return urllib.parse.urljoin(_get_metronome_url(), path)


def _get_timeout():
    """
    :returns: timout value for API calls
    :rtype: str
    """
    # if timeout is not passed, try to read `core.timeout`
    # if `core.timeout` is not set, default to 3 min.
    timeout = config.get_config_val('core.timeout')
    if not timeout:
        timeout = DEFAULT_TIMEOUT

    return timeout



import dcoscli.package.main as package
import dcoscli.util as util


def test_confirm_uninstall_remove_all_fail_challenge():
    real_read = util._read_response

    def fake_read():
        return "hello-world"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', True, '')
    util._read_response = real_read

    assert result is False


def test_confirm_uninstall_remove_all_pass_challenge():
    real_read = util._read_response

    def fake_read():
        return "uninstall all hello-world"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', True, '')
    util._read_response = real_read

    assert result is True


def test_confirm_uninstall_app_id_fail_challenge():
    real_read = util._read_response

    def fake_read():
        return "hello-world"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', False, 'goodbye')
    util._read_response = real_read

    assert result is False


def test_confirm_uninstall_app_id_pass_challenge():
    real_read = util._read_response

    def fake_read():
        return "goodbye"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', False, 'goodbye')
    util._read_response = real_read

    assert result is True


def test_confirm_uninstall_default_fail_challenge():
    real_read = util._read_response

    def fake_read():
        return "kafka"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', False, '')
    util._read_response = real_read

    assert result is False


def test_confirm_uninstall_default_pass_challenge():
    real_read = util._read_response

    def fake_read():
        return "hello-world"

    util._read_response = fake_read
    result = package._confirm_uninstall('hello-world', False, '')
    util._read_response = real_read

    assert result is True



from six.moves import urllib

from dcos import config, http, util
from dcos.errors import (DCOSAuthenticationException,
                         DCOSAuthorizationException,
                         DCOSBadRequest,
                         DCOSException,
                         DCOSHTTPException)

logger = util.get_logger(__name__)


class Cosmos(object):
    """
    A wrapper on cosmos that abstracts away http requests

    :param cosmos_url: the url of cosmos
    :type cosmos_url: str
    """

    def __init__(self, cosmos_url=None):
        if cosmos_url is None:
            self.cosmos_url = get_cosmos_url()
        else:
            self.cosmos_url = cosmos_url

        def _data(versions, http_method):
            """
            Create an object with the data for an endpoint.

            :param versions: the accept versions in order
            of priority.
            :type versions: list[str]
            :param http_method: should be 'post' or 'get'
            :type http_method: str
            :return:
            """
            return {'versions': versions, 'http_method': http_method}

        # This structure holds information about an endpoint. Currently
        # the information stored is the return type versions, and the
        # http request method. These two field should be stored in a
        # dictionary of the form:
        # {'versions': versions, 'http_method': http_method}.
        self._endpoint_data = {
            'capabilities': _data(['v1'], 'get'),
            'package/add': _data(['v1'], 'post'),
            'package/describe': _data(['v3'], 'post'),
            'package/install': _data(['v2'], 'post'),
            'package/list': _data(['v1'], 'post'),
            'package/list-versions': _data(['v1'], 'post'),
            'package/render': _data(['v1'], 'post'),
            'package/repository/add': _data(['v1'], 'post'),
            'package/repository/delete': _data(['v1'], 'post'),
            'package/repository/list': _data(['v1'], 'post'),
            'package/search': _data(['v1'], 'post'),
            'package/uninstall': _data(['v1'], 'post'),
            'service/start': _data(['v1'], 'post')
        }

        self._special_content_types = {
            ('capabilities', 'v1'):
                _format_media_type('capabilities', 'v1', '')
        }

        self._special_accepts = {
            ('capabilities', 'v1'):
                _format_media_type('capabilities', 'v1', '')
        }

    def enabled(self):
        """
        Returns whether or not cosmos is enabled on specified dcos cluster

        :return: true if cosmos is enabled, false otherwise
        :rtype: bool
        """
        try:
            response = self.call_endpoint(
                'capabilities')
        # return `Authentication failed` error messages
        except DCOSAuthenticationException:
            raise
        # Authorization errors mean endpoint exists, and user could be
        # authorized for the command specified, not this endpoint
        except DCOSAuthorizationException:
            return True
        # allow exception through so we can show user actual http exception
        # except 404, because then the url is fine, just not cosmos enabled
        except DCOSHTTPException as e:
            logger.exception(e)
            return e.status() != 404
        except Exception as e:
            logger.exception(e)
            return True
        return response.status_code == 200

    def call_endpoint(self,
                      endpoint,
                      headers=None,
                      data=None,
                      json=None,
                      **kwargs):
        """
        Gets the Response object returned by comos at endpoint

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :param headers: these header values will appear
        in the request headers.
        :type headers: None | dict[str, str]
        :param data: the request's body
        :type data: dict | bytes | file-like object
        :param json: JSON request body
        :type json: dict
        :param kwargs: Additional arguments to requests.request
                       (see py:func:`request`)
        :type kwargs: dict
        :return: the Response object returned by cosmos
        :rtype: requests.Response
        """
        url = self._get_endpoint_url(endpoint)
        request_versions = self._get_request_version_preferences(endpoint)
        headers_preference = list(map(
            lambda version: self._get_header(
                endpoint, version, headers),
            request_versions))
        http_request_type = self._get_http_method(endpoint)
        return self._cosmos_request(
            url,
            http_request_type,
            headers_preference,
            data,
            json,
            **kwargs)

    def _cosmos_request(self,
                        url,
                        http_request_type,
                        headers_preference,
                        data=None,
                        json=None,
                        **kwargs):
        """
        Gets a Response object obtained by calling cosmos
        at the url 'url'. Will attempt each of the headers
        in headers_preference in order until success.

        :param url: the url of a cosmos endpoint
        :type url: str
        :param headers_preference: a list of request headers
        in order of preference. Each header
        will be attempted until they all fail or the request succeeds.
        :type headers_preference: list[dict[str, str]]
        :param data: the request's body
        :type data: dict | bytes | file-like object
        :param json: JSON request body
        :type json: dict
        :param kwargs: Additional arguments to requests.request
                       (see py:func:`request`)
        :type kwargs: dict
        :return: response returned by calling cosmos at url
        :rtype: requests.Response
        """
        try:
            headers = headers_preference[0]
            if http_request_type is 'post':
                response = http.post(
                    url, data=data, json=json, headers=headers, **kwargs)
            else:
                response = http.get(
                    url, data=data, json=json, headers=headers, **kwargs)
            if not _matches_expected_response_header(headers,
                                                     response.headers):
                raise DCOSException(
                    'Server returned incorrect response type, '
                    'expected {} but got {}'.format(
                        headers.get('Accept'),
                        response.headers.get('Content-Type')))
            return response
        except DCOSBadRequest as e:
            if len(headers_preference) > 1:
                # reattempt with one less item in headers_preference
                return self._cosmos_request(url,
                                            http_request_type,
                                            headers_preference[1:],
                                            data,
                                            json,
                                            **kwargs)
            else:
                raise e

    def _get_endpoint_url(self, endpoint):
        """
        Gets the url for the cosmos endpoint 'endpoint'

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :return: the url of endpoint
        :rtype: str
        """
        return urllib.parse.urljoin(self.cosmos_url, endpoint)

    def _get_request_version_preferences(self, endpoint):
        """
        Gets the list of versions for endpoint in preference order.
        The first item is most preferred, and last is least preferred.

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :return: list of versions in preference order
        :rtype: list[str]
        """
        return self._endpoint_data.get(endpoint).get('versions')

    def _get_http_method(self, endpoint):
        """
        Gets the http method cosmos expects for the endpoint
        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :return: http method type
        :rtype: str
        """
        return self._endpoint_data.get(endpoint).get('http_method')

    def _get_header(self, endpoint, version, headers=None):
        """
        Given an cosmos endpoint, a version, and any extra header values,
        gets the header that can be used to query cosmos at endpoint.
        Any key in headers will appear in the final header. In effect the
        user function can overwrite the default header.

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :param version: The version of the request
        :type version: str
        :param headers: extra keys for the header
        :type headers: dict[str, str]
        :return: a header that can be used to query cosmos at endpoint
        :rtype: dict[str, str]
        """
        simple_header = {
            'Content-Type': self._get_content_type(endpoint),
            'Accept': self._get_accept(endpoint, version)
        }
        return _merge_dict(simple_header, headers)

    def _endpoint_exists(self, endpoint):
        """

        :param endpoint: a possible cosmos endpoint
        :type endpoint: str
        :return: true if endpoint is a valid cosmos endpoint,
        false otherwise
        :rtype: bool
        """
        return endpoint in self._endpoint_data

    def _get_accept(self, endpoint, version):
        """
        Gets the value for the Accept header key for
        the cosmos request at endpoint.

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :param version: The version of the request
        :type version: str
        :return: the value for the Accept header key for endpoint
        :rtype: str
        """
        if (endpoint, version) in self._special_accepts:
            return self._special_accepts[(endpoint, version)]
        return _format_media_type(endpoint, version, 'response')

    def _get_content_type(self, endpoint):
        """
        Gets the value for the Content-Type header key for
        the cosmos request at endpoint.

        :param endpoint: a cosmos endpoint, of the form 'x/y',
        for example 'package/repo/add' or 'service/start'
        :type endpoint: str
        :return: the value for the Content-Type header key for endpoint
        :rtype: str
        """
        version = 'v1'
        if (endpoint, version) in self._special_content_types:
            return self._special_content_types[(endpoint, version)]
        return _format_media_type(endpoint, version, 'request')


def _format_media_type(endpoint, version, suffix):
    """
    Formats a value for a cosmos Content-Type or Accept header key.

    :param endpoint: a cosmos endpoint, of the form 'x/y',
    for example 'package/repo/add', 'service/start', or 'package/error'
    :type endpoint: str
    :param version: The version of the request
    :type version: str
    :param suffix: The string that will be appended to
    endpoint type, most commonly 'request' or 'response'
    :type suffix: str
    :return: a formatted value for a Content-Type or Accept header key
    :rtype: str
    """
    prefix = endpoint.replace('/', '.')
    separator = '-' if suffix else ''
    return ('application/vnd.dcos.{}{}{}'
            '+json;charset=utf-8;version={}').format(prefix,
                                                     separator,
                                                     suffix,
                                                     version)


def _matches_expected_response_header(request_headers, response_headers):
    """
    Returns true if the Content-Type value of the response header matches the
    Accept value of the request header, false otherwise

    :param request_headers: the headers for a cosmos request
    :type request_headers: dict[str, str]
    :param response_headers: the headers for a cosmos response
    :type response_headers: dict[str, str]
    :return: true if the Content-Type value of the response header matches the
    Accept value of the request header, false otherwise
    :rtype: bool
    """
    return (request_headers.get('Accept')
            in response_headers.get('Content-Type'))


def get_cosmos_url():
    """
    Gets the cosmos url

    :returns: cosmos base url
    :rtype: str
    """
    toml_config = config.get_config()
    cosmos_url = config.get_config_val('package.cosmos_url', toml_config)
    if cosmos_url is None:
        cosmos_url = config.get_config_val('core.dcos_url', toml_config)
        if cosmos_url is None:
            raise config.missing_config_exception(['core.dcos_url'])
    return cosmos_url


def _merge_dict(a, b):
    """
    Given two dicts, merge them into a new dict as a
    shallow copy. Keys on dictionary b will overwrite keys
    on dictionary a.

    :param a: a dictionary, may be None
    :type a: None | dict
    :param b: a dictionary, may be None
    :type b: None | dict
    :return: the result of merging a with b
    :rtype: dict
    """
    if a is None and b is None:
        return {}

    if a is None:
        return b.copy()

    if b is None:
        return a.copy()

    z = a.copy()
    z.update(b)
    return z



import contextlib
import os
import shutil
import ssl
import urllib

from urllib.request import urlopen

from dcos import config, constants, http, util
from dcos.errors import DCOSException


logger = util.get_logger(__name__)


def move_to_cluster_config():
    """Create a cluster specific config file + directory
    from a global config file. This will move users from global config
    structure (~/.dcos/dcos.toml) to the cluster specific one
    (~/.dcos/clusters/CLUSTER_ID/dcos.toml) and set that cluster as
    the "attached" cluster.

    :rtype: None
    """

    global_config = config.get_global_config()
    dcos_url = config.get_config_val("core.dcos_url", global_config)

    # if no cluster is set, do not move the cluster yet
    if dcos_url is None:
        return

    try:
        # find cluster id
        cluster_url = dcos_url.rstrip('/') + '/metadata'
        res = http.get(cluster_url, timeout=1)
        cluster_id = res.json().get("CLUSTER_ID")

    # don't move cluster if dcos_url is not valid
    except DCOSException as e:
        logger.error(
            "Error trying to find cluster id: {}".format(e))
        return

    # create cluster id dir
    cluster_path = os.path.join(config.get_config_dir_path(),
                                constants.DCOS_CLUSTERS_SUBDIR,
                                cluster_id)

    util.ensure_dir_exists(cluster_path)

    # move config file to new location
    global_config_path = config.get_global_config_path()
    util.sh_copy(global_config_path, cluster_path)

    # set cluster as attached
    util.ensure_file_exists(os.path.join(
        cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE))


@contextlib.contextmanager
def setup_directory():
    """
    A context manager for the temporary setup directory created as a
    placeholder before we find the cluster's CLUSTER_ID.

    :returns: path of setup directory
    :rtype: str
    """

    try:
        temp_path = os.path.join(config.get_config_dir_path(),
                                 constants.DCOS_CLUSTERS_SUBDIR,
                                 "setup")
        util.ensure_dir_exists(temp_path)

        yield temp_path
    finally:
        shutil.rmtree(temp_path, ignore_errors=True)


def setup_cluster_config(dcos_url, temp_path, stored_cert):
    """
    Create a cluster directory for cluster specified in "temp_path"
    directory.

    :param dcos_url: url to DC/OS cluster
    :type dcos_url: str
    :param temp_path: path to temporary config dir
    :type temp_path: str
    :param stored_cert: whether we stored cert bundle in 'setup' dir
    :type stored_cert: bool
    :returns: path to cluster specific directory
    :rtype: str
    """

    try:
        # find cluster id
        cluster_url = dcos_url.rstrip('/') + '/metadata'
        res = http.get(cluster_url, timeout=1)
        cluster_id = res.json().get("CLUSTER_ID")

    except DCOSException as e:
        msg = ("Error trying to find cluster id: {}\n "
               "Please make sure the provided DC/OS URL is valid: {}".format(
                   e, dcos_url))
        raise DCOSException(msg)

    # create cluster id dir
    cluster_path = os.path.join(config.get_config_dir_path(),
                                constants.DCOS_CLUSTERS_SUBDIR,
                                cluster_id)
    if os.path.exists(cluster_path):
        raise DCOSException("Cluster [{}] is already setup".format(dcos_url))

    util.ensure_dir_exists(cluster_path)

    # move contents of setup dir to new location
    for (path, dirnames, filenames) in os.walk(temp_path):
        for f in filenames:
            util.sh_copy(os.path.join(path, f), cluster_path)

    cluster = Cluster(cluster_id)
    config_path = cluster.get_config_path()
    if stored_cert:
        cert_path = os.path.join(cluster_path, "dcos_ca.crt")
        config.set_val("core.ssl_verify", cert_path, config_path=config_path)

    cluster_name = cluster_id
    try:
        url = dcos_url.rstrip('/') + '/mesos/state-summary'
        name_query = http.get(url, timeout=1, toml_config=cluster.get_config())
        cluster_name = name_query.json().get("cluster")

    except DCOSException:
        pass

    config.set_val("cluster.name", cluster_name, config_path=config_path)

    return cluster_path


def set_attached(cluster_path):
    """
    Set the cluster specified in `cluster_path` as the attached cluster

    :param cluster_path: path to cluster directory
    :type cluster_path: str
    :rtype: None
    """

    # get currently attached cluster
    attached_cluster_path = config.get_attached_cluster_path()

    if attached_cluster_path and attached_cluster_path != cluster_path:
        # set cluster as attached
        attached_file = os.path.join(
            attached_cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE)
        try:
            util.sh_move(attached_file, cluster_path)
        except DCOSException as e:
            msg = "Failed to attach cluster: {}".format(e)
            raise DCOSException(msg)

    else:
        util.ensure_file_exists(os.path.join(
            cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE))


def get_cluster_cert(dcos_url):
    """Get CA bundle from specified cluster.

    This is an insecure request.

    :param dcos_url: url to DC/OS cluster
    :type dcos_url: str
    :returns: cert
    :rtype: str
    """

    cert_bundle_url = dcos_url.rstrip() + "/ca/dcos-ca.crt"

    unverified = ssl.create_default_context()
    unverified.check_hostname = False
    unverified.verify_mode = ssl.CERT_NONE

    error_msg = ("Error downloading CA certificate from cluster. "
                 "Please check the provided DC/OS URL.")
    try:
        with urlopen(cert_bundle_url, context=unverified) as f:
            return f.read().decode('utf-8')
    except urllib.error.HTTPError as e:
        # Open source DC/OS does not currently expose its root CA certificate
        if e.code == 404:
            return False
        else:
            logger.debug(e)
            raise DCOSException(error_msg)
    except Exception as e:
        logger.debug(e)
        raise DCOSException(error_msg)


def get_clusters():
    """
    :returns: list of configured Clusters
    :rtype: [Clusters]
    """

    clusters_path = config.get_clusters_path()
    util.ensure_dir_exists(clusters_path)
    clusters = os.listdir(clusters_path)
    return [Cluster(cluster_id) for cluster_id in clusters]


def get_cluster(name):
    """
    :param name: name of cluster
    :type name: str
    :returns: Cluster identified by name
    :rtype: Cluster
    """

    return next((c for c in get_clusters()
                 if c.get_cluster_id() == name or c.get_name() == name), None)


def remove(name):
    """
    Remove cluster `name` from the CLI.

    :param name: name of cluster
    :type name: str
    :rtype: None
    """

    def onerror(func, path, excinfo):
        raise DCOSException("Error trying to remove cluster")

    cluster = get_cluster(name)
    if cluster:
        shutil.rmtree(cluster.get_cluster_path(), onerror)
        return
    else:
        raise DCOSException("Cluster [{}] does not exist".format(name))


class Cluster():
    """Interface for a configured cluster"""

    def __init__(self, cluster_id):
        self.cluster_id = cluster_id
        self.cluster_path = os.path.join(
            config.get_clusters_path(), cluster_id)

    def get_cluster_path(self):
        return self.cluster_path

    def get_cluster_id(self):
        return self.cluster_id

    def get_config_path(self):
        return os.path.join(self.cluster_path, "dcos.toml")

    def get_config(self, mutable=False):
        return config.load_from_path(self.get_config_path(), mutable)

    def get_name(self):
        return config.get_config_val(
            "cluster.name", self.get_config()) or self.cluster_id

    def get_url(self):
        return config.get_config_val("core.dcos_url", self.get_config())

    def get_dcos_version(self):
        dcos_url = self.get_url()
        if dcos_url:
            url = os.path.join(
                self.get_url(), "dcos-metadata/dcos-version.json")
            try:
                resp = http.get(url, timeout=1, toml_config=self.get_config())
                return resp.json().get("version", "N/A")
            except DCOSException:
                pass

        return "N/A"

    def is_attached(self):
        return os.path.exists(os.path.join(
            self.cluster_path, constants.DCOS_CLUSTER_ATTACHED_FILE))

    def __eq__(self, other):
        return isinstance(other, Cluster) and \
            other.get_cluster_id() == self.get_cluster_id()

    def dict(self):
        return {
            "cluster_id": self.get_cluster_id(),
            "name": self.get_name(),
            "url": self.get_url(),
            "version": self.get_dcos_version(),
            "attached": self.is_attached()
        }



import itertools

from dcos import cosmos, emitting, packagemanager, subcommand, util
from dcos.errors import DCOSException

logger = util.get_logger(__name__)

emitter = emitting.FlatEmitter()


def uninstall(pkg, package_name, remove_all, app_id, cli, app):
    """Uninstalls a package.

    :param pkg: package manager to uninstall with
    :type pkg: PackageManager
    :param package_name: The package to uninstall
    :type package_name: str
    :param remove_all: Whether to remove all instances of the named app
    :type remove_all: boolean
    :param app_id: App ID of the app instance to uninstall
    :type app_id: str
    :param init_client: The program to use to run the app
    :type init_client: object
    :rtype: None
    """

    if cli is False and app is False:
        cli = app = True

    uninstalled = False
    installed = installed_packages(
        pkg, app_id, package_name, cli_only=False)
    installed_cli = next((True for installed_pkg in installed
                          if installed_pkg.get("command")), False)
    installed_app = next((True for installed_pkg in installed
                          if installed_pkg.get("apps")), False)

    if cli and installed_cli:
        if subcommand.uninstall(package_name):
            uninstalled = True

    if app and installed_app:
        if pkg.uninstall_app(package_name, remove_all, app_id):
            uninstalled = True

    if uninstalled:
        return None
    else:
        msg = 'Package [{}]'.format(package_name)
        if app_id is not None:
            app_id = util.normalize_marathon_id_path(app_id)
            msg += " with id [{}]".format(app_id)
        msg += " is not installed"
        raise DCOSException(msg)


def uninstall_subcommand(distribution_name):
    """Uninstalls a subcommand.

    :param distribution_name: the name of the package
    :type distribution_name: str
    :returns: True if the subcommand was uninstalled
    :rtype: bool
    """

    return subcommand.uninstall(distribution_name)


def _matches_package_name(name, command_name):
    """
    :param name: the name of the package
    :type name: str
    :param command_name: the name of the command
    :type command_name: str
    :returns: True if the name is not defined or the package matches that name;
              False otherwise
    :rtype: bool
    """

    return name is None or command_name == name


def installed_packages(package_manager, app_id, package_name, cli_only):
    """Returns all installed packages in the format:

    [{
       'apps': [<id>],
       'command': {
         'name': <name>
       }
       ...<metadata>...
    }]

    :param init_client: The program to use to list packages
    :type init_client: object
    :param app_id: App ID of app to show
    :type app_id: str
    :param package_name: The package to show
    :type package_name: str
    :param cli_only: if True, returns only packages with locally installed
        subcommands, without retrieving the apps installed on the cluster
    :type cli_only: bool
    :returns: A list of installed packages matching criteria
    :rtype: [dict]
    """

    apps = []
    if not cli_only:
        apps = package_manager.installed_apps(package_name, app_id)

    subcommands = []
    for subcmd in installed_subcommands():
        if _matches_package_name(package_name, subcmd.name):
            subcmd_dict = subcmd.package_json()
            subcmd_dict['command'] = {'name': subcmd.name}
            subcommands.append(subcmd_dict)

    return merge_installed(apps, subcommands, app_id is not None, cli_only)


def installed_subcommands():
    """Returns all installed subcommands.

    :returns: all installed subcommands
    :rtype: [InstalledSubcommand]
    """

    return [subcommand.InstalledSubcommand(name) for name in
            subcommand.distributions()]


def merge_installed(apps, subcommands, app_only, cli_only):
    """Combines collections of installed apps and subcommands, merging
    elements from the same package.

    :param apps: information on each running app in the cluster; must have
        'name' and 'appId' keys
    :type apps: [dict]
    :param subcommands: information on each subcommand installed locally; must
        have a 'name' key
    :type subcommands: [dict]
    :param app_only: if True, only returns elements that have an app
    :type app_only: bool
    :param cli_only: if True, only returns elements that have a subcommand
    :type cli_only: bool
    :returns: the resulting merged collection, with one element per package
    :rtype: [{}]
    """

    indexed_apps = {}
    grouped_apps = itertools.groupby(apps, key=lambda app: app['name'])
    for name, app_group in grouped_apps:
        app_list = list(app_group)
        pkg = app_list[0]
        pkg['apps'] = sorted(app['appId'] for app in app_list)
        del pkg['appId']
        indexed_apps[name] = pkg

    indexed_subcommands = {subcmd['command']['name']: subcmd
                           for subcmd in subcommands}

    merged = []
    for name, app in indexed_apps.items():
        subcmd = indexed_subcommands.pop(name, {})
        if subcmd or not cli_only:
            app.update(subcmd)
            merged.append(app)

    if not app_only:
        merged.extend(indexed_subcommands.values())

    return merged


def get_package_manager():
    """Returns type of package manager to use

    :returns: PackageManager instance
    :rtype: packagemanager.PackageManager
    """
    cosmos_url = cosmos.get_cosmos_url()
    cosmos_manager = packagemanager.PackageManager(cosmos_url)
    if cosmos_manager.enabled():
        return cosmos_manager
    else:
        msg = ("This version of the DC/OS CLI is not supported for your "
               "cluster. Please downgrade the CLI to an older version: "
               "https://dcos.io/docs/usage/cli/update/#downgrade")
        raise DCOSException(msg)



import requests

from requests.auth import AuthBase

from six.moves.urllib.parse import urlparse

from dcos import config, util
from dcos.errors import (DCOSAuthenticationException,
                         DCOSAuthorizationException, DCOSBadRequest,
                         DCOSConnectionError, DCOSException, DCOSHTTPException,
                         DCOSUnprocessableException)


logger = util.get_logger(__name__)

DEFAULT_TIMEOUT = 5


def _default_is_success(status_code):
    """Returns true if the success status is between [200, 300).

    :param response_status: the http response status
    :type response_status: int
    :returns: True for success status; False otherwise
    :rtype: bool
    """

    return 200 <= status_code < 300


def _verify_ssl(verify=None, toml_config=None):
    """Returns whether to verify ssl

    :param verify: whether to verify SSL certs or path to cert(s)
    :type verify: bool | str
    :param toml_config: cluster config to use
    :type toml_config: Toml
    :return: whether to verify SSL certs or path to cert(s)
    :rtype: bool | str
    """

    if toml_config is None:
        toml_config = config.get_config()

    if verify is None:
        verify = config.get_config_val("core.ssl_verify", toml_config)
        if verify and verify.lower() == "true":
            verify = True
        elif verify and verify.lower() == "false":
            verify = False

    return verify


@util.duration
def _request(method,
             url,
             is_success=_default_is_success,
             timeout=DEFAULT_TIMEOUT,
             auth=None,
             verify=None,
             toml_config=None,
             **kwargs):
    """Sends an HTTP request.

    :param method: method for the new Request object
    :type method: str
    :param url: URL for the new Request object
    :type url: str
    :param is_success: Defines successful status codes for the request
    :type is_success: Function from int to bool
    :param timeout: request timeout
    :type timeout: int
    :param auth: authentication
    :type auth: AuthBase
    :param verify: whether to verify SSL certs or path to cert(s)
    :type verify: bool | str
    :param toml_config: cluster config to use
    :type toml_config: Toml
    :param kwargs: Additional arguments to requests.request
        (see http://docs.python-requests.org/en/latest/api/#requests.request)
    :type kwargs: dict
    :rtype: Response
    """

    if 'headers' not in kwargs:
        kwargs['headers'] = {'Accept': 'application/json'}

    verify = _verify_ssl(verify, toml_config)

    # Silence 'Unverified HTTPS request' and 'SecurityWarning' for bad certs
    if verify is not None:
        silence_requests_warnings()

    logger.info(
        'Sending HTTP [%r] to [%r]: %r',
        method,
        url,
        kwargs.get('headers'))

    try:
        response = requests.request(
            method=method,
            url=url,
            timeout=timeout,
            auth=auth,
            verify=verify,
            **kwargs)
    except requests.exceptions.SSLError as e:
        logger.exception("HTTP SSL Error")
        msg = ("An SSL error occurred. To configure your SSL settings, "
               "please run: `dcos config set core.ssl_verify <value>`")
        description = config.get_property_description("core", "ssl_verify")
        if description is not None:
            msg += "\n<value>: {}".format(description)
        raise DCOSException(msg)
    except requests.exceptions.ConnectionError as e:
        logger.exception("HTTP Connection Error")
        raise DCOSConnectionError(url)
    except requests.exceptions.Timeout as e:
        logger.exception("HTTP Timeout")
        raise DCOSException('Request to URL [{0}] timed out.'.format(url))
    except requests.exceptions.RequestException as e:
        logger.exception("HTTP Exception")
        raise DCOSException('HTTP Exception: {}'.format(e))

    logger.info('Received HTTP response [%r]: %r',
                response.status_code,
                response.headers)

    return response


def request(method,
            url,
            is_success=_default_is_success,
            timeout=None,
            verify=None,
            toml_config=None,
            **kwargs):
    """Sends an HTTP request. If the server responds with a 401, ask the
    user for their credentials, and try request again (up to 3 times).

    :param method: method for the new Request object
    :type method: str
    :param url: URL for the new Request object
    :type url: str
    :param is_success: Defines successful status codes for the request
    :type is_success: Function from int to bool
    :param timeout: request timeout
    :type timeout: int
    :param verify: whether to verify SSL certs or path to cert(s)
    :type verify: bool | str
    :param toml_config: cluster config to use
    :type toml_config: Toml
    :param kwargs: Additional arguments to requests.request
        (see http://docs.python-requests.org/en/latest/api/#requests.request)
    :type kwargs: dict
    :rtype: Response
    """

    if toml_config is None:
        toml_config = config.get_config()

    auth_token = config.get_config_val("core.dcos_acs_token", toml_config)
    prompt_login = config.get_config_val("core.prompt_login", toml_config)
    dcos_url = urlparse(config.get_config_val("core.dcos_url", toml_config))
    cosmos_url = urlparse(
        config.get_config_val("package.cosmos_url", toml_config))
    parsed_url = urlparse(url)

    # only request with DC/OS Auth if request is to DC/OS cluster
    # request should match scheme + netloc
    def _request_match(expected_url, actual_url):
        return expected_url.scheme == actual_url.scheme and \
                    expected_url.netloc == actual_url.netloc

    request_to_cluster = _request_match(dcos_url, parsed_url) or \
        _request_match(cosmos_url, parsed_url)

    if auth_token and request_to_cluster:
        auth = DCOSAcsAuth(auth_token)
    else:
        auth = None

    response = _request(method, url, is_success, timeout,
                        auth=auth, verify=verify, toml_config=toml_config,
                        **kwargs)

    if is_success(response.status_code):
        return response
    elif response.status_code == 401:
        if prompt_login:
            # I don't like having imports that aren't at the top level, but
            # this is to resolve a circular import issue between dcos.http and
            # dcos.auth
            from dcos.auth import header_challenge_auth

            header_challenge_auth(dcos_url.geturl())
            # if header_challenge_auth succeeded, then we auth-ed correctly and
            # thus can safely recursively call ourselves and not have to worry
            # about an infinite loop
            return request(method=method, url=url,
                           is_success=is_success, timeout=timeout,
                           verify=verify, **kwargs)
        else:
            if auth_token is not None:
                msg = ("Your core.dcos_acs_token is invalid. "
                       "Please run: `dcos auth login`")
                raise DCOSAuthenticationException(msg)
            else:
                raise DCOSAuthenticationException(response)
    elif response.status_code == 422:
        raise DCOSUnprocessableException(response)
    elif response.status_code == 403:
        raise DCOSAuthorizationException(response)
    elif response.status_code == 400:
        raise DCOSBadRequest(response)
    else:
        raise DCOSHTTPException(response)


def head(url, **kwargs):
    """Sends a HEAD request.

    :param url: URL for the new Request object
    :type url: str
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('head', url, **kwargs)


def get(url, **kwargs):
    """Sends a GET request.

    :param url: URL for the new Request object
    :type url: str
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('get', url, **kwargs)


def post(url, data=None, json=None, **kwargs):
    """Sends a POST request.

    :param url: URL for the new Request object
    :type url: str
    :param data: Request body
    :type data: dict, bytes, or file-like object
    :param json: JSON request body
    :type data: dict
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('post', url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    """Sends a PUT request.

    :param url: URL for the new Request object
    :type url: str
    :param data: Request body
    :type data: dict, bytes, or file-like object
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('put', url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    """Sends a PATCH request.

    :param url: URL for the new Request object
    :type url: str
    :param data: Request body
    :type data: dict, bytes, or file-like object
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('patch', url, data=data, **kwargs)


def delete(url, **kwargs):
    """Sends a DELETE request.

    :param url: URL for the new Request object
    :type url: str
    :param kwargs: Additional arguments to requests.request
                   (see py:func:`request`)
    :type kwargs: dict
    :rtype: Response
    """

    return request('delete', url, **kwargs)


def silence_requests_warnings():
    """Silence warnings from requests.packages.urllib3.  See DCOS-1007."""
    requests.packages.urllib3.disable_warnings()


class DCOSAcsAuth(AuthBase):
    """Invokes DCOS Authentication flow for given Request object."""
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = "token={}".format(self.token)
        return r



import base64
import json
import os
import shutil
import tempfile
import zipfile

import docopt
import pkg_resources
import six

import dcoscli
from dcos import cmds, emitting, http, options, servicemanager, util
from dcos.errors import DCOSException
from dcos.package import get_package_manager
from dcos.util import md5_hash_file
from dcoscli.subcommand import default_command_info, default_doc
from dcoscli.util import decorate_docopt_usage

logger = util.get_logger(__name__)
emitter = emitting.FlatEmitter()


def main(argv):
    try:
        return _main(argv)
    except DCOSException as e:
        emitter.publish(e)
        return 1


@decorate_docopt_usage
def _main(argv):
    args = docopt.docopt(
        default_doc("experimental"),
        argv=argv,
        version='dcos-experimental version {}'.format(dcoscli.version))
    http.silence_requests_warnings()
    return cmds.execute(_cmds(), args)


def _cmds():
    """
    :returns: All of the supported commands
    :rtype: dcos.cmds.Command
    """
    return [
        cmds.Command(
            hierarchy=['experimental', 'package', 'add'],
            arg_keys=['--json', '--dcos-package',
                      '--package-name', '--package-version'],
            function=_add),
        cmds.Command(
            hierarchy=['experimental', 'package', 'build'],
            arg_keys=['--json', '<build-definition>', '--output-directory'],
            function=_build,
        ),
        cmds.Command(
            hierarchy=['experimental', 'service', 'start'],
            arg_keys=['--json', '<package-name>',
                      '--package-version', '--options'],
            function=_service_start),
        cmds.Command(
            hierarchy=['experimental'],
            arg_keys=['--info'],
            function=_experimental),
    ]


def _experimental(info):
    """
    :returns: process status
    :rtype: int
    """
    if info:
        emitter.publish(default_command_info("experimental"))
    else:
        doc = default_doc("package")
        emitter.publish(options.make_generic_usage_message(doc))
        return 1
    return 0


def _add(json, dcos_package, package_name, package_version):
    """
    Add a DC/OS package to DC/OS

    :param json: wether to output json
    :type json: bool
    :param dcos_package: path to the DC/OS package
    :type dcos_package: None | str
    :param package_name: the name of a remote DC/OS package
    :type package_name: None | str
    :param package_version: the version of a remote DC/OS package
    :type package_version: None | str
    :return: process status
    :rtype: int
    """
    package_manager = get_package_manager()
    if dcos_package:
        response = package_manager.package_add_local(dcos_package)
    else:
        response = (package_manager
                    .package_add_remote(package_name, package_version))

    response_json = response.json()

    if json:
        emitter.publish(response_json)
    else:
        message = (
            'The package [{}] version [{}] has been added to DC/OS'.format(
                response_json['name'], response_json['version']))
        emitter.publish(message)

    return 0


def _build(output_json,
           build_definition,
           output_directory):
    """ Creates a DC/OS Package from a DC/OS Package Build Definition

    :param output_json: whether to output json
    :type output_json: None | bool
    :param build_definition: The Path to a DC/OS package build definition
    :type build_definition: str
    :param output_directory: The directory where the DC/OS Package
    will be stored
    :type output_directory: str
    :returns: The process status
    :rtype: int
    """
    # get the path of the build definition
    cwd = os.getcwd()
    build_definition_path = build_definition
    if not os.path.isabs(build_definition_path):
        build_definition_path = os.path.join(cwd, build_definition_path)

    build_definition_directory = os.path.dirname(build_definition_path)

    if not os.path.exists(build_definition_path):
        raise DCOSException(
            "The file [{}] does not exist".format(build_definition_path))

    # get the path to the output directory
    if output_directory is None:
        output_directory = cwd

    if not os.path.exists(output_directory):
        raise DCOSException(
            "The output directory [{}]"
            " does not exist".format(output_directory))

    logger.debug("Using [%s] as output directory", output_directory)

    # load raw build definition
    with util.open_file(build_definition_path) as bd:
        build_definition_raw = util.load_json(bd, keep_order=True)

    # validate DC/OS Package Build Definition with local references
    build_definition_schema_path = "data/schemas/build-definition-schema.json"
    build_definition_schema = util.load_jsons(
        pkg_resources.resource_string(
            "dcoscli", build_definition_schema_path).decode())

    errs = util.validate_json(build_definition_raw, build_definition_schema)

    if errs:
        logger.debug("Failed before resolution: \n"
                     "\tbuild definition: {}"
                     "".format(build_definition_raw))
        raise DCOSException(_validation_error(build_definition_path))

    # resolve local references in build definition
    _resolve_local_references(
        build_definition_raw,
        build_definition_schema,
        build_definition_directory
    )

    # at this point all the local references have been resolved
    build_definition_resolved = build_definition_raw

    # validate resolved build definition
    metadata_schema_path = "data/schemas/metadata-schema.json"
    metadata_schema = util.load_jsons(
        pkg_resources.resource_string(
            "dcoscli", metadata_schema_path).decode())

    errs = util.validate_json(build_definition_resolved, metadata_schema)

    if errs:
        logger.debug("Failed after resolution: \n"
                     "\tbuild definition: {}"
                     "".format(build_definition_resolved))
        raise DCOSException('Error validating package: '
                            'there was a problem resolving '
                            'the local references in '
                            '[{}]'.format(build_definition_path))

    # create the manifest
    manifest_json = {
        'built-by': "dcoscli.version={}".format(dcoscli.version)
    }

    # create the metadata
    metadata_json = build_definition_resolved

    # create zip file
    with tempfile.NamedTemporaryFile() as temp_file:
        with zipfile.ZipFile(
                temp_file.file,
                mode='w',
                compression=zipfile.ZIP_DEFLATED,
                allowZip64=True) as zip_file:
            metadata = json.dumps(metadata_json, indent=2).encode()
            zip_file.writestr("metadata.json", metadata)

            manifest = json.dumps(manifest_json, indent=2).encode()
            zip_file.writestr("manifest.json", manifest)

        # name the package appropriately
        temp_file.file.seek(0)
        dcos_package_name = '{}-{}-{}.dcos'.format(
            metadata_json['name'],
            metadata_json['version'],
            md5_hash_file(temp_file.file))

        # get the dcos package path
        dcos_package_path = os.path.join(output_directory, dcos_package_name)

        if os.path.exists(dcos_package_path):
            raise DCOSException(
                'Output file [{}] already exists'.format(
                    dcos_package_path))

        # create a new file to contain the package
        temp_file.file.seek(0)
        with util.open_file(dcos_package_path, 'w+b') as dcos_package:
            shutil.copyfileobj(temp_file.file, dcos_package)

    if output_json:
        message = {'package_path': dcos_package_path}
    else:
        message = 'Created DC/OS Universe Package [{}]'.format(
            dcos_package_path)
    emitter.publish(message)

    return 0


def _resolve_local_references(build_definition,
                              build_schema,
                              build_definition_directory):
    """ Resolves all local references in a DC/OS Package Build Definition

    :param build_definition: The DC/OS Package Build Definition that may
     contain local references
    :type build_definition: dict
    :param build_definition_directory: The directory of the Build Definition
    :type build_definition_directory: str
    :param build_schema: The schema for the Build Definition
    :type build_schema: dict
    """
    _replace_marathon(build_definition,
                      build_schema,
                      build_definition_directory)

    _replace_directly(build_definition,
                      build_schema,
                      build_definition_directory,
                      "config")

    _replace_directly(build_definition,
                      build_schema,
                      build_definition_directory,
                      "resource")


def _replace_directly(build_definition,
                      build_schema,
                      build_definition_directory,
                      ref):
    """ Replaces the local reference ref with the contents of
     the file pointed to by ref

    :param build_definition: The DC/OS Package Build Definition that
    may contain local references
    :type build_definition: dict
    :param build_definition_directory: The directory of the Build Definition
    :type build_definition_directory: str
    :param build_schema: The schema for the Build Definition
    :type build_schema: dict
    :param ref: The key in build_definition that will be replaced
    :type ref: str
    """
    if ref in build_definition and _is_local_reference(build_definition[ref]):
        location = build_definition[ref][1:]
        if not os.path.isabs(location):
            location = os.path.join(build_definition_directory, location)

        with util.open_file(location) as f:
            contents = util.load_json(f, True)

        build_definition[ref] = contents

        errs = util.validate_json(build_definition, build_schema)
        if errs:
            logger.debug("Failed during resolution of {}: \n"
                         "\tbuild definition: {}"
                         "".format(ref, build_definition))
            raise DCOSException(_validation_error(location))


def _replace_marathon(build_definition,
                      build_schema,
                      build_definition_directory):
    """ Replaces the marathon v2AppMustacheTemplate ref with
     the base64 encoding of the file pointed to by the reference

    :param build_definition: The DC/OS Package Build Definition that
     may contain local references
    :type build_definition: dict
    :param build_definition_directory: The directory of the Build Definition
    :type build_definition_directory: str
    :param build_schema: The schema for the Build Definition
    :type build_schema: dict
    """
    ref = "marathon"
    template = "v2AppMustacheTemplate"
    if ref in build_definition and \
            _is_local_reference(build_definition[ref][template]):
        location = (build_definition[ref])[template][1:]
        if not os.path.isabs(location):
            location = os.path.join(build_definition_directory, location)

        # convert the contents of the marathon file into base64
        with util.open_file(location) as f:
            contents = base64.b64encode(
                f.read().encode()).decode()

        build_definition[ref][template] = contents

        errs = util.validate_json(build_definition, build_schema)
        if errs:
            logger.debug("Failed during resolution of marathon: \n"
                         "\tbuild definition: {}"
                         "".format(build_definition))
            raise DCOSException(_validation_error(location))


def _validation_error(filename):
    """Renders a human readable validation error

    :param filename: the file that failed to validate
    :type filename: str
    :returns: validation error message
    :rtype: str
    """
    return 'Error validating package: ' \
           '[{}] does not conform to the' \
           ' specified schema'.format(filename)


def _is_local_reference(item):
    """Checks if an object is a local reference

    :param item: the object that may be a reference
    :type item: str
    :returns: true if item is a local reference else false
    :rtype: bool
    """
    return isinstance(item, six.string_types) and item.startswith("@")


def _service_start(json, package_name, package_version, options_path):
    """Starts a DC/OS service from a package that has been added

    :param json: wether to output json
    :type json: bool
    :param package_name:
    :type package_name: str
    :param package_version:
    :type package_version: None | str
    :param options_path:
    :type options_path: None | str
    :return: process status
    :rtype: int
    """
    manager = servicemanager.ServiceManager()
    options = util.read_file_json(options_path) if options_path else None
    response = manager.start_service(
        package_name, package_version, options)
    response_json = response.json()

    if json:
        emitter.publish(response_json)
    else:
        message = (
            "The service [{}] version [{}] has been started".format(
                response_json['packageName'], response_json['packageVersion']
            ))
        emitter.publish(message)

    return 0



from six.moves import urllib

from dcos import config, cosmos
from dcos.errors import (DCOSAuthenticationException,
                         DCOSAuthorizationException,
                         DCOSException,
                         DCOSHTTPException)
from dcos.packagemanager import cosmos_error


class ServiceManager(object):
    """A manager for DC/OS services"""

    def __init__(self, base_url=None):
        self.base_url = base_url if base_url else _get_default_base_url()
        self.cosmos = cosmos.Cosmos(self.base_url)

    def enabled(self):
        """
        Returns whether service manager is enabled.

        :return: true whether this service is enabled, false otherwise
        :rtype: bool
        """
        return self.cosmos.enabled()

    @cosmos_error
    def start_service(self, package_name, package_version, options):
        """
        Starts a service that has been added to the cluster via
        cosmos' package/add endpoint.

        :param package_name: the name of the package to start
        :type package_name: str
        :param package_version: the version of the package to start
        :type package_version: None | str
        :param options: the options for the service
        :type options: None | dict
        :return: the response of cosmos' service/start endpoint
        :rtype: requests.Response
        """
        endpoint = 'service/start'
        json = {'packageName': package_name}
        if package_version is not None:
            json['packageVersion'] = package_version
        if options is not None:
            json['options'] = options
        try:
            return self.cosmos.call_endpoint(endpoint, json=json)
        except (DCOSAuthenticationException, DCOSAuthorizationException):
            raise
        except DCOSHTTPException as e:
            if e.status() == 404:
                message = 'Your version of DC/OS ' \
                          'does not support this operation'
                raise DCOSException(message)
            else:
                return e.response


def _get_default_base_url():
    """
    Gets the default service manager URL

    :returns: cosmos base url
    :rtype: str
    """
    toml_config = config.get_config()
    base_url = config.get_config_val('package.cosmos_url', toml_config)
    if base_url is None:
        base_url = config.get_config_val('core.dcos_url', toml_config)
        if base_url is None:
            raise config.missing_config_exception(['core.dcos_url'])
        else:
            base_url = urllib.parse.urljoin(base_url, 'cosmos/')
    return base_url



from .helpers.common import assert_command


def test_help():
    with open('dcoscli/data/help/help.txt') as content:
        assert_command(['dcos', 'help', '--help'],
                       stdout=content.read().encode('utf-8'))


def test_info():
    assert_command(['dcos', 'help', '--info'],
                   stdout=b'Display help information about DC/OS\n')


def test_version():
    assert_command(['dcos', 'help', '--version'],
                   stdout=b'dcos-help version SNAPSHOT\n')


def test_help_config():
    with open('dcoscli/data/help/config.txt') as content:
        assert_command(['dcos', 'help', 'config'],
                       stdout=content.read().encode('utf-8'))


def test_help_job():
    with open('dcoscli/data/help/job.txt') as content:
        assert_command(['dcos', 'help', 'job'],
                       stdout=content.read().encode('utf-8'))


def test_help_marathon():
    with open('dcoscli/data/help/marathon.txt') as content:
        assert_command(['dcos', 'help', 'marathon'],
                       stdout=content.read().encode('utf-8'))


def test_help_node():
    with open('dcoscli/data/help/node.txt') as content:
        assert_command(['dcos', 'help', 'node'],
                       stdout=content.read().encode('utf-8'))


def test_help_package():
    with open('dcoscli/data/help/package.txt') as content:
        assert_command(['dcos', 'help', 'package'],
                       stdout=content.read().encode('utf-8'))


def test_help_service():
    with open('dcoscli/data/help/service.txt') as content:
        assert_command(['dcos', 'help', 'service'],
                       stdout=content.read().encode('utf-8'))


def test_help_task():
    with open('dcoscli/data/help/task.txt') as content:
        assert_command(['dcos', 'help', 'task'],
                       stdout=content.read().encode('utf-8'))


def test_help_auth():
    with open('dcoscli/data/help/auth.txt') as content:
        assert_command(['dcos', 'help', 'auth'],
                       stdout=content.read().encode('utf-8'))
